"""
Load database from SNAP format.
"""
import json, glob, os
import torch
import numpy as np

from .database import Database
from .restarter import Restartable
from ..tools import pad_np_array_to_length_with_zeros as padax
from ..tools import progress_bar, np_of_torchdefaultdtype

from ase.data import chemical_symbols


class SNAPDirectoryDatabase(Database, Restartable):
    def __init__(
        self,
        directory,
        inputs,
        targets,
        *args,
        files=None,
        depth=1,
        transpose_cell=True,
        allow_unfound=False,
        quiet=False,
        **kwargs,
    ):

        self.directory = directory
        self.files = files
        self.inputs = inputs
        self.targets = targets
        self.transpose_cell = transpose_cell
        self.depth = depth
        arr_dict = self.load_arrays(quiet=quiet, allow_unfound=allow_unfound)

        super().__init__(arr_dict, inputs, targets, *args, **kwargs, allow_unfound=allow_unfound, quiet=quiet)

        self.restarter = self.make_restarter(
            directory,
            inputs,
            targets,
            *args,
            transpose_cell=transpose_cell,
            files=files,
            allow_unfound=allow_unfound,
            **kwargs,
            quiet=quiet,
        )

    def load_arrays(self, allow_unfound=False, quiet=False):

        if not self.files:
            if not quiet:
                print("Acquiring file list")
            glob_str = f"{self.directory}" + ("/*" * self.depth) + "/*.json"
            files = glob.glob(glob_str)
            if len(files) == 0:
                raise FileNotFoundError(f"No '.json' files found in directory {self.directory}")
        else:
            files = [os.path.join(self.directory, f) for f in self.files]

        files.sort()

        config_unprocessed = []
        for f in progress_bar(files, desc="Data Files", unit="file"):
            config_unprocessed.append(self.extract_snap_file(f))

        config_unprocessed = [c for batch in config_unprocessed for c in batch]  # Flattening groups

        n_atoms_max = max(d["NumAtoms"] for d in config_unprocessed)

        arr_dict = self.process_configs(config_unprocessed, n_atoms_max)
        arr_dict = self.filter_arrays(arr_dict, allow_unfound=allow_unfound, quiet=quiet)
        return arr_dict

    def filter_arrays(self, arr_dict, allow_unfound=False, quiet=False):
        if not quiet:
            print("Arrays found: ", list(arr_dict.keys()))

        floatX = np_of_torchdefaultdtype()
        for k, v in arr_dict.copy().items():
            if not allow_unfound:
                if k not in self.inputs and k not in self.targets:
                    del arr_dict[k]
            if v.dtype == "float64":
                arr_dict[k] = v.astype(floatX)

        if not quiet:
            print("Data types:")
            print({k: v.dtype for k, v in arr_dict.items()})

        return arr_dict

    def extract_snap_file(self, file):
        with open(file, "rt") as jf:
            comment = jf.readline()
            content = jf.read()
        parsed = json.loads(content)
        dataset = parsed["Dataset"]
        data_items = dataset["Data"]

        for i, d in enumerate(data_items):
            group_path, config_name = os.path.split(file)
            base_path, group_name = os.path.split(file)
            data_items[i]["FileName"] = config_name
            data_items[i]["Group"] = group_name
            data_items[i]["SubConfig"] = i

        return data_items

    def process_configs(self, configs, n_atoms_max):

        arr_dict = {}
        all_keys = "AtomTypes", "Energy", "Forces", "Lattice", "Positions", "Group", "FileName", "SubConfig"
        pad_keys = "AtomTypes", "Forces", "Positions"
        for key in all_keys:
            value_list = [c[key] for c in configs]
            if key in pad_keys:
                value_list = [padax(np.asarray(v), n_atoms_max) for v in value_list]
            arr_dict[key] = np.stack(value_list)
        arr_dict["AtomTypes"][arr_dict["AtomTypes"] == "0"] = "X"  # ASE calls blank atoms 'X'
        z_array = [[chemical_symbols.index(s) for s in sym] for sym in arr_dict["AtomTypes"]]
        arr_dict["Species"] = np.asarray(z_array).astype(int)
        arr_dict["AtomCount"] = arr_dict["Species"].astype(bool).sum(axis=1)
        arr_dict["EnergyPerAtom"] = arr_dict["Energy"] / arr_dict["AtomCount"]

        if self.transpose_cell:
            arr_dict["Lattice"] = arr_dict["Lattice"].transpose((0, 2, 1))

        return arr_dict



"""
Package for tracking node index state.
The index state describes the meaning of the
batch axis or axes for the tensor. i.e. system level, atom level, pair level.
"""
import warnings

from ... import settings

if settings.DEBUG_AUTOINDEXING:
    warnings.warn("Printing automatic index coercion info! Output will be verbose.")


def _debprint(*args, **kwargs):
    if settings.DEBUG_AUTOINDEXING:
        print("AutoIndex:", *args, **kwargs)
    else:
        pass


from .type_def import IdxType
from .registry import register_index_transformer, clear_index_cache
from .reduce_funcs import (
    elementwise_compare_reduce,
    get_reduced_index_state,
    index_type_coercion,
    soft_index_type_coercion,
    db_form,
)

__all__ = [
    "IdxType",
    "register_index_transformer",
    "clear_index_cache",
    "elementwise_compare_reduce",
    "get_reduced_index_state",
    "index_type_coercion",
    "soft_index_type_coercion",
    "db_form",
]



import torch
from . import indexers
from torch import Tensor


class NACR(torch.nn.Module):
    """
    Compute NAC vector * ΔE. Originally in hippynn.layers.physics.
    """

    def __init__(self):
        super().__init__()

    def forward(
        self,
        charges1: Tensor,
        charges2: Tensor,
        positions: Tensor,
        energy1: Tensor,
        energy2: Tensor,
    ):
        dE = energy2 - energy1
        nacr = torch.autograd.grad(
            charges2, [positions], grad_outputs=[charges1], create_graph=True
        )[0].reshape(len(dE), -1)
        return nacr * dE


class NACRMultiState(torch.nn.Module):
    """
    Compute NAC vector * ΔE for all paris of states. Originally in hippynn.layers.physics.
    """

    def __init__(self, n_target=1):
        self.n_target = n_target
        super().__init__()

    def forward(self, charges: Tensor, positions: Tensor, energies: Tensor):
        # charges shape: n_molecules, n_atoms, n_targets
        # positions shape: n_molecules, n_atoms, 3
        # energies shape: n_molecules, n_targets
        # dE shape: n_molecules, n_targets, n_targets
        dE = energies.unsqueeze(1) - energies.unsqueeze(2)
        # take the upper triangle excluding the diagonal
        indices = torch.triu_indices(
            self.n_target, self.n_target, offset=1, device=dE.device
        )
        # dE shape: n_molecules, n_pairs
        # n_pairs = n_targets * (n_targets - 1) / 2
        dE = dE[..., indices[0], indices[1]]
        # compute q1 * dq2/dR
        nacr_ij = []
        for i, j in zip(*indices):
            nacr = torch.autograd.grad(
                charges[..., j],
                positions,
                grad_outputs=charges[..., i],
                create_graph=True,
            )[0]
            nacr_ij.append(nacr)
        # nacr shape: n_molecules, n_atoms, 3, n_pairs
        nacr = torch.stack(nacr_ij, dim=1)
        n_molecule, n_pairs, n_atoms, n_dims = nacr.shape
        nacr = nacr.reshape(n_molecule, n_pairs, n_atoms * n_dims)
        # multiply dE
        return nacr * dE.unsqueeze(2)


class LocalEnergy(torch.nn.Module):
    def __init__(self, feature_sizes, first_is_interacting=False):

        super().__init__()
        self.first_is_interacting = first_is_interacting
        if first_is_interacting:
            feature_sizes = feature_sizes[1:]

        self.feature_sizes = feature_sizes

        self.summer = indexers.MolSummer()
        self.n_terms = len(feature_sizes)
        biases = (first_is_interacting, *(True for _ in range(self.n_terms - 1)))

        self.layers = torch.nn.ModuleList(torch.nn.Linear(nf, 1, bias=bias) for nf, bias in zip(feature_sizes, biases))
        self.players = torch.nn.ModuleList(torch.nn.Linear(nf, 1, bias=False) for nf in feature_sizes)
        self.ninf = float("-inf")

    def forward(self, all_features, mol_index, atom_index, n_molecules, n_atoms_max):
        """
        :param all_features: list of feature tensors
        :param mol_index: which molecule is the atom
        :param atom_index: which atom in the molecule is that atom
        :param n_molecules: total number of molecules in the batch
        :param n_atoms_max: maximum number of atoms in the batch
        :return: contributed_energy, atom_energy, atom_preenergy, prob, propensity
        """

        if self.first_is_interacting:
            all_features = all_features[1:]

        partial_preenergy = [lay(x) for x, lay in zip(all_features, self.layers)]
        atom_preenergy = sum(partial_preenergy)
        partial_potentials = [lay(x) for x, lay in zip(all_features, self.players)]
        propensity = sum(partial_potentials)  # Keep in mind that this has shape (natoms,1)

        # This segment does not need gradients, we are constructing the subtraction parameters for softmax
        # which results in a calculation that does not under or overflow; the result is most accurate this way
        # But actually invariant to the subtraction used, so it does not require a grad.
        # It's a standard SoftMax technique, however, the implementation is not built into pytorch for
        # the molecule/atom framework.
        with torch.autograd.no_grad():
            propensity_molatom = all_features[0].new_full((n_molecules, n_atoms_max, 1), self.ninf)
            propensity_molatom[mol_index, atom_index] = propensity
            propensity_norms = propensity_molatom.max(dim=1)[0]  # first element is max vals, 2nd is max position
            propensity_norm_atoms = propensity_norms[mol_index]

        propensity_normed = propensity - propensity_norm_atoms

        # Calculate probabilities with molecule version of softmax
        relative_prob = torch.exp(propensity_normed)
        z_factor_permol = self.summer(relative_prob, mol_index, n_molecules)
        atom_zfactor = z_factor_permol[mol_index]
        prob = relative_prob / atom_zfactor

        # Find molecular sum
        atom_energy = prob * atom_preenergy
        contributed_energy = self.summer(atom_energy, mol_index, n_molecules)

        return contributed_energy, atom_energy, atom_preenergy, prob, propensity




import torch

from .open import _PairIndexer

class FilterDistance(_PairIndexer):
    """ Filters a list of tensors in *pair_lists by distance. 
    pair_dist is first positional argument.

    :param _PairIndexer: FilterDistance subclasses _PairIndexer so that the
    FilterPairIndexers behave as regular PairIndexers. 
    """
    
    def forward(self, pair_dist, *pair_tensors):
        r_cut = self.hard_dist_cutoff 
        idx = torch.argwhere(pair_dist <= r_cut).squeeze(1)
        pair_tensors = pair_dist, *pair_tensors
        return tuple(pl[idx] for pl in pair_tensors)




"""
Predictor object which builds a graph for predictions with
a simple user interface.
"""
import torch

from ..tools import arrdict_len, progress_bar

from .graph import GraphModule
from .gops import search_by_name
from .indextypes import IdxType, db_form

# TODO: batched prediction for scalars
# It doesn't work for scalars! Right now they are just filtered out.


class Predictor:
    """
    The predictor is a dressed-up GraphModule which gives access
    to the outputs of individual nodes.

    In many cases you may simply want to use the ``from_graph`` method to generate a predictor.

    The predictor will take the model graph, convert the output nodes into
    a padded index state, and build a new graph for these operations.

    """

    def __init__(self, inputs, outputs, return_device=torch.device("cpu"), model_device=None, requires_grad=False,
                 name=None):
        """

        :param inputs: nodes to use as inputs
        :param outputs: nodes to use as outputs
        :param return_device: device to place results on; does nothing if None
        :param model_device: where the model and input should be placed, does nothing if None
        :param requires_grad: Default false -- detach predicted tensors. If true, do not detach predicted
         tensors -- this may lead to memory leaks if not done carefully.
        :param name: Optional name for formatting progress bars.

        """

        outputs = [search_by_name(inputs, o) if isinstance(o, str) else o for o in outputs]
        outputs = list(set(outputs))  # Remove any redundancies -- they will screw up the output name map.

        outputs = [o for o in outputs if o._index_state is not IdxType.Scalar]

        self.out_names = [o.name for o in outputs]
        self.out_dbnames = [o.db_name for o in outputs]

        outputs = [db_form(o) for o in outputs]
        self.graph = GraphModule(inputs, outputs)

        # User can supply kwargs using the node's name or db_name, the latter takes priority
        self.argmap = {
            **{node.name: node for node in self.graph.input_nodes},
            **{node.db_name: node for node in self.graph.input_nodes},
        }

        self.return_device = return_device
        self.model_device = model_device
        self.requires_grad = requires_grad
        self.name = name or "Predictor"

    @classmethod
    def from_graph(cls, graph, additional_outputs=None, **kwargs):
        """
        Construct a new predictor from an existing GraphModule.

        :param graph:  graph to create predictor for. The predictor makes a shallow copy of this graph. e.g. it may move
           parameters from that graph to the model_device.
        :param additional_outputs: List of additional nodes to include in outputs
        :param kwargs: passed to ``__init__``

        :return: predictor instance
        """
        inputs = graph.input_nodes
        outputs = graph.nodes_to_compute
        if additional_outputs is not None:
            outputs = outputs + list(additional_outputs)

        return cls(inputs, outputs, **kwargs)

    def to(self, *args, **kwargs):
        return self.graph.to(*args, **kwargs)

    @property
    def inputs(self):
        return self.graph.input_nodes

    @property
    def outputs(self):
        return self.graph.nodes_to_compute

    @property
    def model_device(self):
        return self._model_device

    @model_device.setter
    def model_device(self, device):
        self.graph.to(device)
        self._model_device = device

    def add_output(self, node):
        if isinstance(node, str):
            node = self.graph.node_from_name(node)

        source = node
        node = db_form(node)
        if node not in self.graph.forward_output_list:
            raise ValueError(
                "Node {} is not computed in graph. We suggest you make a ".format(node)
                + "new predictor with this node in the additional_outputs list."
            )

        self.outputs.append(node)
        self.out_names.append(source.name)
        self.out_dbnames.append(source.bname)

    def wrap_outputs(self, out_dict):
        for (node, tensor), dbname, name in zip(out_dict.copy().items(), self.out_dbnames, self.out_names):
            out_dict[name] = tensor
            out_dict[dbname] = tensor
        return out_dict

    def __call__(self, *, node_values=None, single_prediction=False, batch_size=None, **kwargs):
        """

        :param node_values: dict[node->tensor]
        :param single_prediction: (default false) if True, wrap a batch axis onto the input and off of the output
        :param batch_size: batch over groups of examples by splitting the inputs along the batch axis.
                           does all computation as one batch if batch_size=None
        :param kwargs: values for arguments, using either the input's name or dbname attribute
        :return: dictionary of output values. Contains db_name:value, name:value, and node:value pairs.
        """

        if node_values is not None:
            if len(kwargs) != 0:
                raise ValueError("Cannot call Predictor with both node_values and kwargs")
            pass
        else:
            node_values = {self.argmap[k]: v for k, v in kwargs.items()}

        if len(node_values) != len(self.inputs):
            raise ValueError(
                "Wrong number of node inputs. {} distinct inputs supplied, "
                " {} inputs required".format(len(node_values), len(self.inputs))
            )

        if single_prediction:
            node_values = {n: v.unsqueeze(0) for n, v in node_values.items()}

        if batch_size is not None:
            out = self.predict_batched(node_values, batch_size)
        else:
            out = self.predict_all(node_values)

        if single_prediction:
            out = {n: v.unsqueeze(0) for n, v in out.items()}

        out = self.wrap_outputs(out)

        return out

    def predict_batched(self, node_values, batch_size):

        # Split the input into batches
        node_values = {n: torch.split(t, batch_size) for n, t in node_values.items()}

        n_batches = arrdict_len(node_values)

        # Make a dictionary for each batch
        batched_inputs = [{n: t[i] for n, t in node_values.items()} for i in range(n_batches)]

        # Predict for each batch
        batched_outputs = [self.predict_all(b) for b in progress_bar(batched_inputs, desc=self.name, unit="batch")]

        # Concatenate the full set of outputs.
        out_keys = batched_outputs[0].keys()
        full_outputs = {k: torch.cat([batched_outputs[i][k] for i in range(n_batches)]) for k in out_keys}
        return full_outputs

    def apply_to_database(self, db, **kwargs):
        """
        Note: kwargs are passed to self.__call__, e.g. the ``batch_size`` parameter.
        """
        results = {}
        for split_name, split_arrdict in db.splits.items():
            input_names = [x.db_name for x in self.graph.input_nodes]
            dict_inputs = {k: split_arrdict[k] for k in input_names}
            results[split_name] = self(**dict_inputs, **kwargs)
        return results

    def predict_all(self, node_values):

        input_values = [node_values[n] for n in self.inputs]

        if self.model_device is not None:
            input_values = [inp.to(self.model_device) for inp in input_values]

        out_values = self.graph(*input_values)

        out_dict = {k: v for k, v in zip(self.outputs, out_values)}

        if not self.requires_grad:
            out_dict = {k: v.detach() for k, v in out_dict.items()}

        if self.return_device is not None:
            out_dict = {k: v.to(self.return_device) for k, v in out_dict.items()}

        return out_dict



"""
Nodes for finding and manipulating pairs and distances.
"""

from .base.node_functions import NodeNotFound
from .base import AutoNoKw, AutoKw, ExpandParents, SingleNode, MultiNode, find_unique_relative, _BaseNode
from .indexers import PaddingIndexer, acquire_encoding_padding, OneHotEncoder
from .tags import Encoder, PairIndexer, AtomIndexer, PairCache
from .inputs import PositionsNode, CellNode, SpeciesNode
from ..indextypes import IdxType
from ...layers import pairs as pairs_modules


class OpenPairIndexer(ExpandParents, PairIndexer, MultiNode):
    _input_names = "coordinates", "nonblank", "real_atoms", "inv_real_atoms"
    _auto_module_class = pairs_modules.OpenPairIndexer

    @_parent_expander.match(PositionsNode, SpeciesNode)
    def expand0(self, pos, spec, *, purpose, **kwargs):
        enc = find_unique_relative(spec, Encoder, why_desc=purpose)
        padidx = find_unique_relative(spec, PaddingIndexer, why_desc=purpose)
        return pos, enc, padidx

    @_parent_expander.match(PositionsNode, Encoder, PaddingIndexer)
    def expand0(self, pos, encode, indexer, **kwargs):
        return pos, encode.nonblank, indexer.real_atoms, indexer.inv_real_atoms

    _parent_expander.assertlen(4)
    _parent_expander.get_main_outputs()
    _parent_expander.require_idx_states(IdxType.MolAtom, None, None, None)

    def __init__(self, name, parents, dist_hard_max, module="auto", **kwargs):
        self.dist_hard_max = dist_hard_max
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)

    def auto_module(self):
        return self._auto_module_class(self.dist_hard_max)


class PeriodicPairOutputs:
    _output_names = "pair_dist", "pair_first", "pair_second", "pair_coord", "cell_offsets", "offset_index"
    _output_index_states = (IdxType.Pair,) * len(_output_names)


class PeriodicPairIndexer(ExpandParents, AutoKw, PeriodicPairOutputs, PairIndexer, MultiNode):
    _input_names = "coordinates", "nonblank", "real_atoms", "inv_real_atoms", "cell"
    _auto_module_class = pairs_modules.PeriodicPairIndexer

    @_parent_expander.match(PositionsNode, SpeciesNode, CellNode)
    def expand0(self, pos, spec, cell, *, purpose, **kwargs):
        enc, padidx = acquire_encoding_padding(spec, species_set=None)
        return pos, enc, padidx, cell

    @_parent_expander.match(PositionsNode, Encoder, PaddingIndexer, CellNode)
    def expand1(self, pos, encode, indexer, cell, **kwargs):
        return pos, encode.nonblank, indexer.real_atoms, indexer.inv_real_atoms, cell

    _parent_expander.assertlen(5)
    _parent_expander.get_main_outputs()

    def __init__(self, name, parents, dist_hard_max, module="auto", module_kwargs=None, **kwargs):
        if module_kwargs is None:
            module_kwargs = {}
        self.module_kwargs = {"hard_dist_cutoff": dist_hard_max, **module_kwargs}
        self.dist_hard_max = self.module_kwargs["hard_dist_cutoff"]

        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)

class Memory:
    @property
    def skin(self):
        return self.torch_module.skin
    
    @skin.setter
    def skin(self, skin):
        self.torch_module.skin = skin

    @property
    def reuse_percentage(self):
        return self.torch_module.reuse_percentage
    
    def reset_reuse_percentage(self):
        self.torch_module.reset_reuse_percentage()

class PeriodicPairIndexerMemory(PeriodicPairIndexer, Memory):
    '''
    Implementation of PeriodicPairIndexer with additional memory component.

    Stores current pair indices in memory and reuses them to compute the pair distances if no 
    particle has moved more than skin/2 since last pair calculation. Otherwise uses the
    _pair_indexer_class to recompute the pairs.

    Increasing the value of 'skin' will increase the number of pair distances computed at
    each step, but decrease the number of times new pairs must be computed. Skin should be 
    set to zero while training for fastest results.
    '''
    
    _auto_module_class = pairs_modules.periodic.PeriodicPairIndexerMemory

    def __init__(self, name, parents, dist_hard_max, skin, module="auto", module_kwargs=None, **kwargs):
        if module_kwargs is None:
            module_kwargs = {}
        self.module_kwargs = {"skin": skin, **module_kwargs}

        super().__init__(name, parents, dist_hard_max, module=module, module_kwargs=self.module_kwargs, **kwargs)


class ExternalNeighborIndexer(ExpandParents, PairIndexer, AutoKw, MultiNode):
    _input_names = "coordinates", "real_atoms", "shifts", "cell", "ext_pair_first", "ext_pair_second"
    _auto_module_class = pairs_modules.ExternalNeighbors

    _parent_expander.get_main_outputs()
    _parent_expander.assertlen(len(_input_names))
    _parent_expander.require_idx_states(IdxType.MolAtom, IdxType.MolAtom, None, None, None, None)

    def __init__(self, name, parents, hard_dist_cutoff, module="auto", **kwargs):
        self.module_kwargs = {"hard_dist_cutoff": hard_dist_cutoff}
        super().__init__(name, parents, module=module, **kwargs)


# Pair reindexer to re-use existing pairs
class PairReIndexer(ExpandParents, AutoNoKw, SingleNode):
    """
    For re-using index information to convert
    from IdxType.MolAtomAtom -> IdxType.Pairs
    """

    _input_names = "pair_features", "molecule_index", "atom_index", "pair_first", "pair_second"
    _auto_module_class = pairs_modules.PairReIndexer
    _index_state = IdxType.Pair

    @_parent_expander.match(_BaseNode)
    def expand0(self, pair_features):
        pad_idx = find_unique_relative(pair_features, PaddingIndexer)
        pair_idx = find_unique_relative(pair_features, PairIndexer)
        return pair_features, pad_idx, pair_idx

    @_parent_expander.match(_BaseNode, PaddingIndexer, PairIndexer)
    def expand1(self, pair_features, pad_idx, pair_idx):
        return (
            pair_features.main_output,
            pad_idx.molecule_index,
            pad_idx.atom_index,
            pair_idx.pair_first,
            pair_idx.pair_second,
        )

    _parent_expander.assertlen(5)
    _parent_expander.get_main_outputs()

    def __init__(self, name, parents, module="auto", **kwargs):
        super().__init__(name, parents, module=module, **kwargs)


# Pair deindexer to convert pair features back to padded form


class PairDeIndexer(ExpandParents, AutoNoKw, SingleNode):
    """
    For converting from IdxType.Pairs to IdxType.MolAtomAtom
    (Padded form)
    """

    _input_names = (
        "pair_features",
        "molecule_index",
        "atom_index",
        "n_molecules",
        "n_atoms_max" "pair_first",
        "pair_second",
    )
    _auto_module_class = pairs_modules.PairDeIndexer
    _index_state = IdxType.MolAtomAtom

    @_parent_expander.match(_BaseNode)
    def expand0(self, pair_features):
        pad_idx = find_unique_relative(pair_features, PaddingIndexer)
        pair_idx = find_unique_relative(pair_features, PairIndexer)
        return pair_features, pad_idx, pair_idx

    @_parent_expander.match(_BaseNode, PaddingIndexer, PairIndexer)
    def expand1(self, pair_features, pad_idx, pair_idx):
        return (
            pair_features.main_output,
            pad_idx.molecule_index,
            pad_idx.atom_index,
            pad_idx.n_molecules,
            pad_idx.n_atoms_max,
            pair_idx.pair_first,
            pair_idx.pair_second,
        )

    def __init__(self, name, parents, module="auto", **kwargs):
        super().__init__(name, parents, module=module, **kwargs)


class PairCacher(ExpandParents, AutoKw, PairCache, SingleNode):
    _input_names = (
        "pair_first",
        "pair_second",
        "cell_offsets",
        "offset_index",
        "real_atoms",
        "mol_index",
        "n_atoms_max",
        "n_molecules",
    )
    _auto_module_class = pairs_modules.PairCacher
    _index_state = IdxType.NotFound

    @_parent_expander.match(PairIndexer)
    def expand0(self, pair_indexer, *args, purpose, **kwargs):
        atomidx = find_unique_relative(pair_indexer, AtomIndexer)
        if "n_images" not in self.module_kwargs:
            self.module_kwargs["n_images"] = pair_indexer.torch_module.n_images
        return pair_indexer, atomidx

    @_parent_expander.match(PairIndexer, AtomIndexer)
    def expand1(self, pair_indexer, atomidx, *args, purpose, **kwargs):
        mi = atomidx.mol_index
        nam = atomidx.n_atoms_max
        n_molecules = atomidx.n_molecules
        ra = atomidx.real_atoms
        pf = pair_indexer.pair_first
        ps = pair_indexer.pair_second
        po = pair_indexer.cell_offsets
        poi = pair_indexer.offset_index
        return pf, ps, po, poi, ra, mi, n_molecules, nam

    _parent_expander.assertlen(8)
    _parent_expander.require_idx_states(IdxType.Pair, IdxType.Pair, None, None, None, None, None, None)

    def __init__(self, name, parents, module="auto", module_kwargs=None, **kwargs):
        self.module_kwargs = module_kwargs or {}
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)


class PairUncacher(ExpandParents, AutoNoKw, PairIndexer, MultiNode):
    _input_names = "sparsepairs", "coordinates", "cells", "real_atoms", "inv_real_atoms", "n_atoms_max", "n_molecules"
    _output_names = "pair_dist", "pair_first", "pair_second", "pair_coord", "cell_offsets", "offset_index"
    _output_index_states = (IdxType.Pair,) * len(_output_names)
    _auto_module_class = pairs_modules.PairUncacher
    _index_state = IdxType.NotFound

    @_parent_expander.match(PairCache)
    def expand0(self, sparse, *args, purpose, **kwargs):
        pos = find_unique_relative(sparse, PositionsNode)
        cell = find_unique_relative(sparse, CellNode)
        atomidx = find_unique_relative(sparse, AtomIndexer)
        return sparse, pos, cell, atomidx

    @_parent_expander.match(PairCache, PositionsNode, CellNode, AtomIndexer)
    def expand1(self, sp, r, c, atomidx, *args, purpose, **kwargs):
        ira = atomidx.inv_real_atoms
        nam = atomidx.n_atoms_max
        n_molecules = atomidx.n_molecules
        ra = atomidx.real_atoms
        return sp, r, c, ra, ira, nam, n_molecules

    _parent_expander.assertlen(7)

    def __init__(self, name, parents, dist_hard_max, module="auto", **kwargs):
        self.dist_hard_max = dist_hard_max
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)


class RDFBins(ExpandParents, AutoKw, SingleNode):
    _input_names = "pair_dists", "pair_first", "pair_second", "one_hot", "n_molecules"
    _index_state = None
    _auto_module_class = pairs_modules.RDFBins

    @_parent_expander.match(PositionsNode, SpeciesNode, CellNode)
    def expand0(self, pos, spec, cell, *, purpose, dist_hard_max=None, **kwargs):
        """
        Build a default Periodic Pair indexer.
        """
        pairs = PeriodicPairIndexer("Period Pairs", (pos, spec, cell), dist_hard_max=dist_hard_max)
        return pairs

    @_parent_expander.match(PositionsNode, SpeciesNode)
    def expand1(self, pos, spec, *, purpose, dist_hard_max=None, **kwargs):
        """
        Builds an open pair indexer.
        """
        pairs = OpenPairIndexer("Period Pairs", (pos, spec), dist_hard_max=dist_hard_max)
        return pairs

    @_parent_expander.match(PairIndexer)
    def expand2(self, pairs, *, purpose, **kwargs):
        """
        Get the encoding and padding associated with a pair indexer.
        """
        enc = find_unique_relative(pairs, OneHotEncoder)
        pad = find_unique_relative(pairs, PaddingIndexer)
        return pairs, enc, pad

    @_parent_expander.match(PairIndexer, OneHotEncoder, PaddingIndexer)
    def expand3(self, pairs, one_hot, pad, *, purpose, **kwargs):
        """
        Expanded the needed children of pairs, encoder, and padding indexer.
        """
        self.module_kwargs["species_set"] = one_hot.species_set
        return pairs.pair_dist, pairs.pair_first, pairs.pair_second, one_hot.encoding, pad.n_molecules

    _parent_expander.require_idx_states(IdxType.Pair, IdxType.Pair, IdxType.Pair, IdxType.Atoms, None)
    _parent_expander.assertlen(5)

    def __init__(self, name, parents, module="auto", bins=None, module_kwargs=None, **kwargs):
        if module_kwargs is None:
            module_kwargs = {}
        self.module_kwargs = {"bins": bins, **module_kwargs}
        parents = self.expand_parents(parents, dist_hard_max=max(bins))
        super().__init__(name, parents, module=module, **kwargs)


class _DispatchNeighbors(ExpandParents, AutoKw, PeriodicPairOutputs, PairIndexer, MultiNode):
    """
    Superclass for nodes that compute neighbors for systems one at a time.
    These should be capable of searching all feasible neighbors (no limit on number of images)
    """

    _input_names = (
        "coordinates",
        "nonblank",
        "real_atoms",
        "inv_real_atoms",
        "cell",
        "mol_index",
        "n_molecules",
        "n_atoms_max",
    )
    # Needs auto_module_class or explicit module

    @_parent_expander.match(PositionsNode, SpeciesNode, CellNode)
    def expand0(self, pos, spec, cell, **kwargs):
        """
        Acquire padding and encoding.
        """
        enc, padidx = acquire_encoding_padding(spec, species_set=None)
        return pos, enc, padidx, cell

    @_parent_expander.match(PositionsNode, Encoder, PaddingIndexer, CellNode)
    def expand1(self, pos, encode, indexer, cell, **kwargs):
        """
        Expand needed child nodes of encoder and padding indexer.
        """

        return (
            pos,
            encode.nonblank,
            indexer.real_atoms,
            indexer.inv_real_atoms,
            cell,
            indexer.mol_index,
            indexer.n_molecules,
            indexer.n_atoms_max,
        )

    _parent_expander.assertlen(8)
    _parent_expander.get_main_outputs()
    _parent_expander.require_idx_states(IdxType.MolAtom, None, None, None, None, None, None, None)

    def __init__(self, name, parents, dist_hard_max, module="auto", module_kwargs=None, **kwargs):
        self.dist_hard_max = dist_hard_max
        parents = self.expand_parents(parents)

        if module_kwargs is None:
            module_kwargs = {}
        self.module_kwargs = {"dist_hard_max": dist_hard_max, **module_kwargs}

        super().__init__(name, parents, module=module, **kwargs)


class NumpyDynamicPairs(_DispatchNeighbors):
    _auto_module_class = pairs_modules.NPNeighbors


class DynamicPeriodicPairs(_DispatchNeighbors):
    """
    Node for finding pairs in arbitrary periodic boundary conditions.
    Note: This will often be slower than PeriodicPairIndexer, but more general.
    If the speed is a concern, consider precomputing pairs with experiment.assembly.precompute_pairs
    """

    _auto_module_class = pairs_modules.TorchNeighbors

class KDTreePairs(_DispatchNeighbors):
    '''
    Node for finding pairs under periodic boundary conditions using Scipy's KD Tree algorithm. 
    Cell must be orthorhombic.
    '''
    _auto_module_class = pairs_modules.dispatch.KDTreeNeighbors

class KDTreePairsMemory(_DispatchNeighbors, Memory):
    '''
    Implementation of KDTreePairs with an added memory component.

    Stores current pair indices in memory and reuses them to compute the pair distances if no 
    particle has moved more than skin/2 since last pair calculation. Otherwise uses the
    _pair_indexer_class to recompute the pairs.

    Increasing the value of 'skin' will increase the number of pair distances computed at
    each step, but decrease the number of times new pairs must be computed. Skin should be 
    set to zero while training for fastest results.
    '''
    _auto_module_class = pairs_modules.dispatch.KDTreePairsMemory

    def __init__(self, name, parents, dist_hard_max, skin, module="auto", module_kwargs=None, **kwargs):
        if module_kwargs is None:
            module_kwargs = {}
        module_kwargs = {"skin": skin, **module_kwargs}

        super().__init__(name, parents, dist_hard_max, module=module, module_kwargs=module_kwargs, **kwargs)

class PaddedNeighborNode(ExpandParents, AutoNoKw, MultiNode):
    _input_names = "pair_first", "pair_second", "pair_coord"
    _output_names = (
        "j_list",
        "rij_list",
    )
    _output_index_states = IdxType.Atoms, IdxType.Atoms
    _auto_module_class = pairs_modules.PaddedNeighModule

    @_parent_expander.match(PairIndexer)
    def expand0(self, pair_finder, **kwargs):
        try:
            # Typically, the first atom tensor will come from
            # the output of the atom indexer, so look for that first.
            pad = pair_finder.find_unique_relative(AtomIndexer)
            atom_array = pad.indexed_features
        except NodeNotFound:
            # Fall back to finding -any- atom-indexed tensor.
            atom_arrays = pair_finder.find_relatives(
                lambda node: hasattr(node, "_index_state") and node._index_state == IdxType.Atoms
            )
            atom_array = atom_arrays.pop()

        return pair_finder.pair_first, pair_finder.pair_second, pair_finder.pair_coord, atom_array

    _parent_expander.assertlen(4)
    _parent_expander.get_main_outputs()
    _parent_expander.require_idx_states(IdxType.Pair, IdxType.Pair, IdxType.Pair, IdxType.Atoms)

    def __init__(self, name, parents, module="auto", **kwargs):
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)


class MinDistNode(ExpandParents, AutoNoKw, MultiNode):
    _input_names = "rij_list", "j_list", "mol_index", "atom_index", "inv_real_atoms", "n_atoms_max", "n_molecules"
    _output_names = "min_dist_mol", "mol_locs", "min_dist_atom", "atom_pairlocs"
    _output_index_states = IdxType.Molecules, IdxType.Molecules, IdxType.Atoms, IdxType.Atoms
    _auto_module_class = pairs_modules.MinDistModule

    @_parent_expander.match(PairIndexer)
    def expand0(self, pair_finder, **kwargs):

        try:
            neigh_list = pair_finder.find_unique_relative(PaddedNeighborNode)
        except NodeNotFound:
            neigh_list = PaddedNeighborNode("NeighList", pair_finder)

        return (neigh_list,)

    @_parent_expander.match(PaddedNeighborNode)
    def expand1(self, neigh_list, **kwargs):
        pad = neigh_list.find_unique_relative(AtomIndexer)
        return neigh_list, pad

    @_parent_expander.match(PaddedNeighborNode, AtomIndexer)
    def expand2(self, neigh_list, pad_idxer, **kwargs):
        return (
            neigh_list.rij_list,
            neigh_list.j_list,
            pad_idxer.mol_index,
            pad_idxer.atom_index,
            pad_idxer.inv_real_atoms,
            pad_idxer.n_atoms_max,
            pad_idxer.n_molecules,
        )

    _parent_expander.assertlen(7)
    _parent_expander.get_main_outputs()
    _parent_expander.require_idx_states(IdxType.Atoms, IdxType.Atoms, None, None, None, None, None)

    def __init__(self, name, parents, module="auto", **kwargs):
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)


# Graph Nodes for Filter Pair Indexer. Works with PeriodicPairs, OpenPairs, External Neighbors.
class PairFilter(AutoKw, PairIndexer, ExpandParents, MultiNode):
    _auto_module_class = pairs_modules.FilterDistance

    @_parent_expander.match(PairIndexer)
    def expand0(self, pair_indexer, purpose):

        # During graph construction, every node is connected to its current set of parents. 
        # It is possible that pair_indexer.children can contain itself; an un-initialized PairFilter. 
        # Only initialized PairIndexers are extracted here. 
        parents = [c for c in pair_indexer.children if hasattr(c, "_index_state")]

        # Validate that nothing unexpected has happened.
        # Hopefully this can't fail, but if we update the pair API or someone customizes this aspect of the
        # library, this should catch any problems.
        idx_states = set(c._index_state for c in parents)

        if len(idx_states) != 1:
            raise TypeError(f"Input contains mixed index states: {idx_states}. Input states should only consist of index type pair.")
        idx_state = idx_states.pop()
        if idx_state != IdxType.Pair:
            raise TypeError(f"Index state for inputs was {idx_state}, needs to be index type pair.")
        # Validation complete.
        self._output_names = tuple(f"out_{name}" for name in pair_indexer._output_names)
        self._input_names = tuple(f"in_{name}" for name in pair_indexer._output_names)
        self._output_index_states = (IdxType.Pair,)*len(parents)

        return parents

    def __init__(self, name, parents, dist_hard_max, module="auto", **kwargs):
        self.module_kwargs = {"hard_dist_cutoff": dist_hard_max}  # passes to PairIndexer superclass
        self.dist_hard_max = dist_hard_max
        parents = self.expand_parents(parents)
        super().__init__(name, parents, module=module, **kwargs)



"""
Implementations of pair finding, pair index manipulation, and analysis of pairs
"""

# Possible TODO do periodic coordinates ever need to be differentiable with respect to the cell
# For volume changes?

from .open import OpenPairIndexer, _PairIndexer

from .periodic import PeriodicPairIndexer

from .filters import FilterDistance

from .indexing import (
    ExternalNeighbors,
    PairDeIndexer,
    PairReIndexer,
    PairCacher,
    PairUncacher,
    MolPairSummer,
    PaddedNeighModule,
)

from .analysis import RDFBins, MinDistModule

from .dispatch import NPNeighbors, TorchNeighbors



"""
Core GraphModule that implements node computations
"""
import warnings
from collections import OrderedDict

import torch

from .nodes.base.node_functions import _BaseNode
from .nodes.base import InputNode

from . import get_subgraph, compute_evaluation_order

from .. import settings


# possible TODO: Enable "virtual input nodes" where we accept the value for any given node
# Requires copying subgraph, eliminating the parents which belong only to the
# Virtual input nodes, and then creating an input node for each virtual input node
# Idea: allow someone to evaluate parts of the graph when other parts are known
# implementation: use swap_parent and generate input nodes with the right index
# type.


class GraphModule(torch.nn.Module):
    def __init__(self, required_inputs, nodes_to_compute):
        """

        takes a graph, makes a pytorch neural network

        :param required_inputs: inputs
        :param nodes_to_compute: outouts
        """
        super().__init__()

        assert all(isinstance(n, InputNode) for n in required_inputs)

        self.input_nodes = required_inputs
        self.nodes_to_compute = nodes_to_compute

        all_node_list = get_subgraph(nodes_to_compute)

        for node in all_node_list:
            if isinstance(node, InputNode):
                if node not in self.input_nodes:
                    raise ValueError("Nodes to compute requires an unspecified input:", node)

        for x in all_node_list:
            if not isinstance(x, _BaseNode):
                if not callable(x):
                    raise ValueError("Computational objects must be callable.")
                else:
                    warnings.warn("Warning: object not base node:", x)

        del x

        self.forward_output_list, self.forward_inputs_list = compute_evaluation_order(all_node_list)

        # Here we have to do a small dance to allow us to create a dictionary of nodes to modules.
        # Pytorch requires ModuleDicts to have keys that are strings.
        # We want a ModuleDict with keys that are nodes.
        # self.node_mod_map does this as a regular python dictionary,
        # while registration with pytorch is given by numbering.

        self.names_dict = OrderedDict((n, "node" + str(i)) for i, n in enumerate(self.forward_output_list))
        self.moddict = torch.nn.ModuleDict(
            OrderedDict(
                (
                    (node_name, node.torch_module)
                    for node, node_name in self.names_dict.items()
                    if not isinstance(node, InputNode)
                )
            )
        )

    def get_module(self, node):
        return self.moddict[self.names_dict[node]]

    def print_structure(self):
        """Pretty-print the structure of the nodes and links comprising this graph."""
        in_nodes = {n: "I{}".format(i) for i, n in enumerate(self.input_nodes)}
        out_nodes = {n: "O{}".format(i) for i, n in enumerate(self.nodes_to_compute)}
        middle_nodes = {n: "H{}".format(i) for i, n in enumerate(self.forward_output_list)}
        middle_nodes.update(out_nodes)
        node_map = {**in_nodes, **middle_nodes}
        print("Inputs:")
        for k, v in in_nodes.items():
            print("\t", v, ":", k)
        print("Outputs:")
        for k, v in out_nodes.items():
            print("\t", v, ":", k)
        print("Order:")
        for computed, inputs_for_computed in zip(self.forward_output_list, self.forward_inputs_list):
            pre = ",".join({node_map[n] for n in inputs_for_computed})
            mid = "{:3} : {}".format(node_map[computed], computed.name)
            print("{:-<20}-> {}".format(pre, mid))

        node_map.update()
        node_map.update()
        node_map.update({n: "Out{}:".format(i) for i, n in enumerate(self.forward_output_list)})

    def node_from_name(self, name):
        for match in list(self.input_nodes) + list(self.forward_output_list):
            if match.db_name == name or match.name == name:
                node = match
                break
        else:
            raise ValueError("Name '{}' not found in graph.".format(name))
        return node

    def extra_repr(self):
        return "Inputs: {} \n Outputs: {}".format(
            tuple(x.name for x in self.input_nodes), tuple(x.name for x in self.nodes_to_compute)
        )

    def forward(self, *input_values):
        # Add gradient computation if needed (e.g. for force computation)
        computed = {
            node: value.requires_grad_(value.requires_grad or node.requires_grad)
            for node, value in zip(self.input_nodes, input_values)
        }

        for this_node, inputs_for_this_node in zip(self.forward_output_list, self.forward_inputs_list):
            computed[this_node] = self.get_module(this_node)(*(computed[inkey] for inkey in inputs_for_this_node))

        return tuple(computed[x] for x in self.nodes_to_compute)


class _DebugGraphModule(GraphModule):
    def forward(self, *input_values):
        try:
            computed = {
                node: value.requires_grad_(value.requires_grad or node.requires_grad)
                for node, value in zip(self.input_nodes, input_values)
            }
        except Exception as ee:
            for n, v in zip(self.input_nodes, input_values):
                print(
                    "\tNode:",
                    n,
                )
                print("\t\tValue:", v)
            raise Exception("Something broke loading the values.") from ee

        print("INPUTTED:")
        for k, v in computed.items():
            print("\t", k.name, v.device, v.shape, v)
        print("Should be inputted:")
        print(*(i.name for i in self.input_nodes), sep="\n\t")

        for this_node, inputs_for_this_node in zip(self.forward_output_list, self.forward_inputs_list):
            print("Computing:", this_node.name)
            try:
                inputs = tuple(computed[inkey] for inkey in inputs_for_this_node)
                module = self.get_module(this_node)
                print("module:", module)
                print("Inputs for this node:", [x.name for x in inputs_for_this_node])
                print("input devices:", [(x.device if isinstance(x, torch.Tensor) else type(x)) for x in inputs])
                value = module(*inputs)
                # print("inputs:",inputs)
                # print("Value:",value)
                if isinstance(value, torch.Tensor):
                    print("Output shape:", value.shape, value.dtype)
                    if torch.isnan(value).any():
                        raise ValueError("Nan detected! {}".format(value))
                else:
                    print("Output non-tensor.")
                computed[this_node] = value
            except Exception as e:
                print("Problem!")
                print("Inputs for this node:", [x.name for x in inputs_for_this_node])
                for in_node in inputs_for_this_node:
                    in_comp = computed[in_node]
                    if isinstance(in_comp, torch.Tensor):
                        print("\t", in_node.name, in_comp.dtype, in_comp.shape)
                    else:
                        print("\t non-tensor:", in_comp)

                print(e)
                print("Computing state:")
                print({k.name: k for k in computed.keys()})
                # for k,v in computed.items():
                #    print("Name: {} Value: {}".format(k.name,v))
                raise e

        return tuple(computed[x] for x in self.nodes_to_compute)


if settings.DEBUG_GRAPH_EXECUTION:
    GraphModule = _DebugGraphModule



"""
Core classes and functions for nodes
"""
from .node_functions import find_unique_relative, find_relatives, get_connected_nodes

# Basic node classes
from .base import Node, SingleNode, InputNode, LossInputNode, LossPredNode, LossTrueNode, _BaseNode

# Node that provides multiple outputs
from .multi import MultiNode

# Optional mixins for simplifying the process of defining BaseNode subclasses
from .definition_helpers import AutoKw, AutoNoKw, ExpandParents



import ase
import torch

from ...graphs.nodes.pairs import _DispatchNeighbors
from ...layers import pairs as pair_modules


def ASE_compute_neighbors(cutoff, positions, cell):
    positions = positions.detach().cpu().numpy()
    cell = cell.detach().cpu().numpy()

    nlist = ase.neighborlist.NeighborList(
        cutoff,
        skin=0.0,
        sorted=True,
        self_interaction=False,
        bothways=True,
        primitive=ase.neighborlist.NewPrimitiveNeighborList,
    )
    n_atoms = len(positions)
    symbols = ["H"] * n_atoms  # We don't actually care what atom types we use

    atoms = ase.Atoms(symbols=symbols, positions=positions, cell=cell, pbc=True)
    nlist.update(atoms)
    pf = nlist.nl.pair_first
    ps = nlist.nl.pair_second
    offset = -nlist.nl.offset_vec  # Different conventions in ASE!
    return tuple(torch.from_numpy(x) for x in [pf, ps, offset])


class ASENeighbors(pair_modules.dispatch._DispatchNeighbors):
    def compute_one(self, positions, cell):
        return ASE_compute_neighbors(self.dist_hard_max, positions, cell)


class ASEPairNode(_DispatchNeighbors):
    _auto_module_class = ASENeighbors


from ...experiment import assembly

assembly._PAIRCACHE_COMPATIBLE_COMPUTERS.add(ASEPairNode)



"""
Base database functionality from dictionary of numpy arrays
"""
import warnings
import numpy as np
import torch

from .restarter import NoRestart
from ..tools import arrdict_len, device_fallback

from torch.utils.data import DataLoader, TensorDataset, Subset


class Database:
    """
    Class for holding a pytorch dataset, splitting it, generating dataloaders, etc."
    """

    def __init__(
        self,
        arr_dict,
        inputs,
        targets,
        seed,
        test_size=None,
        valid_size=None,
        num_workers=0,
        pin_memory=True,
        allow_unfound=False,
        quiet=False,
    ):
        """
        :param arr_dict: dictionary mapping strings to numpy arrays
        :param inputs:   list of strings for input db_names
        :param targets:  list of strings for output db_namees
        :param seed:     int, for random splitting
        :param test_size: fraction of data to use in test split
        :param valid_size: fraction of data to use in train split
        :param num_workers: passed to pytorch dataloaders
        :param pin_memory: passed to pytorch dataloaders
        :param allow_unfound: If true, skip checking if the needed inputs and targets are found.
           This allows setting inputs=None and/or targets=None.
        :param quiet: If True, print little or nothing while loading.
        """

        # Restartable Children of this class should change this after super().__init__ .
        self.restarter = NoRestart()

        self.inputs = inputs
        self.targets = targets
        self.quiet = quiet
        self.splitting_completed = False
        self.num_workers = num_workers
        self.pin_memory = pin_memory

        if not quiet:
            print(f"All arrays:")
            prettyprint_arrays(arr_dict)

        if not allow_unfound:
            for k in self.var_list:
                if k not in arr_dict and k not in ("indices", "split_indices"):
                    raise KeyError(
                        f"Array dictionary missing required variable:'{k}'."
                        "Pass allow_unfound=True to avoid checking of inputs targets."
                    )

        self.arr_dict = arr_dict
        if "indices" not in arr_dict:
            if not quiet:
                print("Database: Using auto-generated data indices")
            self.arr_dict["indices"] = np.arange(len(self), dtype=int)
        else:
            if not quiet:
                print("Database: Using pre-specified data indices.")

        self.splits = {}
        self.random_state = np.random.RandomState(seed=seed)

        if test_size is not None or valid_size is not None:
            if test_size is None or valid_size is None:
                raise ValueError("Both test and valid size must be set for auto-splitting")
            else:
                self.make_trainvalidtest_split(test_size=test_size, valid_size=valid_size)

    def __len__(self):
        return arrdict_len(self.arr_dict)

    @property
    def var_list(self):
        if self.inputs is None:
            raise RuntimeError(f"Database inputs not defined, set {Database}.inputs.")
        if self.targets is None:
            raise RuntimeError(f"Database inputs not defined, set {Database}.targets.")
        return self.inputs + self.targets

    def send_to_device(self, device=None):
        "Send a database to a device"
        if len(self.splits) == 0:
            raise RuntimeError("Arrays must be split before sending database to device.")
        if device is None:
            device = device_fallback()
        else:
            device = torch.device(device)
        if device.type != "cpu":
            if self.pin_memory:
                warnings.warn("Pin memory was set, but target device requested is not CPU. Setting pin_memory=False.")
                self.pin_memory = False
            if self.num_workers != 0:
                warnings.warn("Num workers was set, but target device requested is not CPU. Setting num_workers=0.")
                self.num_workers = 0

        for split, arrdict in self.splits.items():
            for k in arrdict:
                arrdict[k] = arrdict[k].to(device)

    def make_random_split(self, evaluation_mode, split_size):
        """

        :param evaluation_mode: String naming the split, can be anything, but 'train', 'valid', and 'test' are special.s
        :param split_size: int (number of items) or float<1, fraction of samples.
        :return:
        """
        if self.splitting_completed:
            raise RuntimeError("Database already split!")

        if split_size < 1:
            split_size = int(split_size * len(self))

        split_indices = self.random_state.choice(self.arr_dict["indices"], size=split_size, replace=False)

        split_indices.sort()

        return self.make_explicit_split(evaluation_mode, split_indices)

    def make_trainvalidtest_split(self, test_size, valid_size):
        if self.splitting_completed:
            raise RuntimeError("Database already split!")

        if valid_size < 1:
            if test_size >= 1:
                raise ValueError("If train or valid size is set as a fraction, then set test_size as a fraction")
            else:
                if valid_size + test_size > 1:
                    raise ValueError(
                        f"Test fraction ({test_size}) plus valid fraction " f"({valid_size}) are greater than 1!"
                    )
                valid_size /= 1 - test_size

        self.make_random_split("test", test_size)
        self.make_random_split("valid", valid_size)
        self.split_the_rest("train")

    def make_explicit_split(self, evaluation_mode, split_indices):
        if self.splitting_completed:
            raise RuntimeError("Database already split!")

        if len(split_indices) == 0:
            raise ValueError("Cannot make split of size 0.")
        # Compute which indices are not being split off.
        index_mask = compute_index_mask(split_indices, self.arr_dict["indices"])

        complement_mask = ~index_mask

        # Split off data, and keep the rest.
        self.splits[evaluation_mode] = {k: torch.from_numpy(self.arr_dict[k][index_mask]) for k in self.arr_dict}
        self.splits[evaluation_mode]["split_indices"] = torch.arange(len(split_indices), dtype=torch.int64)

        for k, v in self.arr_dict.items():
            self.arr_dict[k] = v[complement_mask]

        if not self.quiet:
            print(f"Arrays for split: {evaluation_mode}")
            prettyprint_arrays(self.splits[evaluation_mode])

    def split_the_rest(self, evaluation_mode):
        self.make_explicit_split(evaluation_mode, self.arr_dict["indices"])
        self.splitting_completed = True

    def make_generator(self, split_type, evaluation_mode, batch_size=None, subsample=False):
        """
        Makes a dataloader for the given type of split and evaluation mode of the model.

        :param split_type: str; "train", "valid", or "test" ; selects data to use
        :param evaluation_mode: str; "train" or "eval". Used for whether to shuffle.
        :param batch_size: passed to pytorch
        :param subsample: fraction to subsample
        :return: dataloader containing relevant data

        """
        if not self.splitting_completed:
            raise ValueError("Database has not yet been split.")

        if split_type not in self.splits:
            raise ValueError(f"Split {split_type} Invalid. Current splits:{list(self.splits.keys())}")

        data = [self.splits[split_type][k] for k in self.var_list]

        if evaluation_mode == "train":
            if split_type != "train":
                raise ValueError(
                    "evaluation mode 'train' can only be used with training data." "(got {})".format(split_type)
                )
            shuffle = True
        elif evaluation_mode == "eval":
            shuffle = False
        else:
            raise ValueError(f"Evaluation_mode ({evaluation_mode}) must be one of 'train' or 'eval'")

        dataset = NamedTensorDataset(self.var_list, *data)
        if subsample:
            n_total = data[0].shape[0]
            n_selected = int(n_total * subsample)
            sampled_indices = torch.argsort(torch.rand(n_total))[:n_selected]
            # sampled_indices = torch.rand(data[0].shape[0]) < subsample
            dataset = Subset(dataset, sampled_indices)
            # data = [a[sampled_indices] for a in data]

        generator = DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            pin_memory=self.pin_memory,
            num_workers=self.num_workers,
        )

        return generator
        
    def trim_all_arrays(self,index):
        """
        To be used in conjuction with remove_high_property
        """
        for key in self.arr_dict:
            self.arr_dict[key] = self.arr_dict[key][index]
    
    def remove_high_property(self,key,perAtom,species_key=None,cut=None,std_factor=10):
        """
        This function removes outlier data from the dataset
        Must be called before splitting
        "key": the property key in the dataset to check for high values
        "perAtom": True if the property is defined per atom in axis 1, otherwise property is treated as full system
        "std_factor": systems with values larger than this multiplier time the standard deviation of all data will be reomved. None to skip this step
        "cut_factor": systems with values larger than this number are reomved. None to skip this step. This step is done first. 
        """
        if perAtom:
            if species_key==None:
                raise RuntimeError("species_key must be defined to trim a per atom quantity")
            atom_ind = self.arr_dict[species_key] > 0
        ndim = len(self.arr_dict[key].shape)
        if cut!=None:
            if perAtom:
                Kmean = np.mean(self.arr_dict[key][atom_ind])
            else:
                Kmean = np.mean(self.arr_dict[key])
            failArr = np.abs(self.arr_dict[key]-Kmean)>cut
            #This does nothing with ndim=1
            trimArr = np.sum(failArr,axis=tuple(range(1,ndim)))==0
            self.trim_all_arrays(trimArr)
            
        if std_factor!=None:
            if perAtom:
                atom_ind = self.arr_dict[species_key] > 0
                Kmean = np.mean(self.arr_dict[key][atom_ind])
                std_cut = np.std(self.arr_dict[key][atom_ind]) * std_factor
            else: 
                Kmean = np.mean(self.arr_dict[key])
                std_cut = np.std(self.arr_dict[key]) * std_factor
            failArr = np.abs(self.arr_dict[key]-Kmean)>std_cut
            #This does nothing with ndim=1
            trimArr = np.sum(failArr,axis=tuple(range(1,ndim)))==0
            self.trim_all_arrays(trimArr)



def compute_index_mask(indices, index_pool):
    if not np.all(np.isin(indices, index_pool)):
        raise ValueError("Provided indices not in database")

    uniques, counts = np.unique(indices, return_counts=True)
    if len(uniques) != len(indices):
        raise ValueError("Split indices not unique")
    if counts.max() > 1:
        raise ValueError("Split indices have duplicates.")

    index_mask = np.isin(index_pool, indices)
    return index_mask


def prettyprint_arrays(arr_dict):
    """
    Pretty-print array dictionary
    :return: None
    """
    column_format = "| {:<18} | {:<18} | {:<40} |"
    ncols = len(column_format.format("", "", ""))

    def printrow(*args):
        print(column_format.format(*args))

    def printline():
        print("-" * ncols)

    printline()
    printrow("Name", "dtype", "shape")
    printline()
    for key, value in arr_dict.items():
        printrow(key, repr(value.dtype), repr(value.shape))
    printline()


class NamedTensorDataset(TensorDataset):
    def __init__(self, tensor_names, *tensors):
        super().__init__(*tensors)
        self.tensor_map = tensor_names



"""
Layers for HIP-NN
"""
import numpy as np
import torch
import warnings
from .. import custom_kernels

from .. import settings


def warn_if_under(distance, threshold):
    if len(distance) == 0:  # no pairs
        return
    dmin = distance.min()
    if dmin < threshold:
        d_count = distance < threshold
        d_frac = d_count.to(distance.dtype).mean()
        d_sum = (d_count.sum()/2).to(torch.int)
        warnings.warn(
            "Provided distances are underneath sensitivity range!\n"
            f"Minimum distance in current batch: {dmin}\n"
            f"Threshold distance for warning: {threshold}.\n"
            f"Fraction of pairs under the threshold: {d_frac}\n"
            f"Number of pairs under the threshold: {d_sum}"
        )


class CosCutoff(torch.nn.Module):
    def __init__(self, hard_max_dist):
        super().__init__()
        self.hard_max_dist = hard_max_dist

    def forward(self, dist_tensor):
        cutoff_sense = torch.cos(np.pi / 2 * dist_tensor / self.hard_max_dist) ** 2
        cutoff_sense = cutoff_sense * (dist_tensor <= self.hard_max_dist).to(cutoff_sense.dtype)
        return cutoff_sense


class SensitivityModule(torch.nn.Module):
    def __init__(self, hard_max_dist, cutoff_type):
        super().__init__()
        self.cutoff = cutoff_type(hard_max_dist)
        self.hard_max_dist = hard_max_dist


class GaussianSensitivityModule(SensitivityModule):
    def __init__(self, n_dist, min_dist_soft, max_dist_soft, hard_max_dist, cutoff_type=CosCutoff):

        super().__init__(hard_max_dist, cutoff_type)
        init_mu = 1.0 / torch.linspace(1.0 / max_dist_soft, 1.0 / min_dist_soft, n_dist)
        self.mu = torch.nn.Parameter(init_mu.unsqueeze(0))

        self.sigma = torch.nn.Parameter(torch.Tensor(n_dist).unsqueeze(0))
        init_sigma = min_dist_soft * 2 * n_dist  # pulled from theano code
        self.sigma.data.fill_(init_sigma)

    def forward(self, distflat, warn_low_distances=None):
        if warn_low_distances is None:
            warn_low_distances = settings.WARN_LOW_DISTANCES
        if warn_low_distances:
            with torch.no_grad():
                mu, argmin = self.mu.min(dim=1)
                sig = self.sigma[:, argmin]
                # Warn if distance is less than the -inside- edge of the shortest sensitivity function
                thresh = mu + sig
                warn_if_under(distflat, thresh)
        distflat_ds = distflat.unsqueeze(1)
        mu_ds = self.mu
        sig_ds = self.sigma

        nondim = (distflat_ds**-1 - mu_ds**-1) ** 2 / (sig_ds**-2)
        base_sense = torch.exp(-0.5 * nondim)

        total_sense = base_sense * self.cutoff(distflat).unsqueeze(1)
        return total_sense


class InverseSensitivityModule(SensitivityModule):
    def __init__(self, n_dist, min_dist_soft, max_dist_soft, hard_max_dist, cutoff_type=CosCutoff):

        super().__init__(hard_max_dist, cutoff_type)
        init_mu = torch.Tensor(1.0 / torch.linspace(1.0 / max_dist_soft, 1.0 / min_dist_soft, n_dist))
        self.mu = torch.nn.Parameter(init_mu.unsqueeze(0))
        self.sigma = torch.nn.Parameter(torch.Tensor(n_dist).unsqueeze(0))
        init_sigma = min_dist_soft * 2 * n_dist
        self.sigma.data.fill_(init_sigma)

    def forward(self, distflat, warn_low_distances=None):
        if warn_low_distances is None:
            warn_low_distances = settings.WARN_LOW_DISTANCES
        if warn_low_distances:
            with torch.no_grad():
                # Warn if distance is less than the -inside- edge of the shortest sensitivity function
                mu, argmin = self.mu.min(dim=1)
                sig = self.sigma[:, argmin]
                thresh = (mu**-1 - sig**-1) ** -1

                warn_if_under(distflat, thresh)
        distflat_ds = distflat.unsqueeze(1)

        nondim = (distflat_ds**-1 - self.mu**-1) ** 2 / (self.sigma**-2)
        base_sense = torch.exp(-0.5 * nondim)

        total_sense = base_sense * self.cutoff(distflat).unsqueeze(1)

        return total_sense


class SensitivityBottleneck(torch.nn.Module):
    def __init__(
        self,
        n_dist,
        min_soft_dist,
        max_dist_soft,
        hard_max_dist,
        n_dist_bare,
        cutoff_type=CosCutoff,
        base_sense=InverseSensitivityModule,
    ):
        super().__init__()
        self.hard_max_dist = hard_max_dist

        self.base_sense = base_sense(n_dist_bare, min_soft_dist, max_dist_soft, hard_max_dist, cutoff_type)
        self.matching = torch.nn.Parameter(torch.Tensor(n_dist_bare, n_dist))

        self.cutoff = self.base_sense.cutoff

        torch.nn.init.orthogonal_(self.matching.data)

    def forward(self, distflat):
        base_sense = self.base_sense(distflat)
        reduced_sense = torch.mm(base_sense, self.matching)
        return reduced_sense


class InteractLayer(torch.nn.Module):
    """
    Hipnn's interaction layer
    """

    def __init__(self, nf_in, nf_out, n_dist, mind_soft, maxd_soft, hard_cutoff, sensitivity_module):
        """
        Constructor

        :param nf_in: number of input features
        :param nf_out: number of output features
        :param n_dist: number of distance sensitivities
        :param mind_soft: minimum distance for initial sensitivities
        :param maxd_soft: maximum distance for initial sensitivities
        :param hard_cutoff: maximum distance for cutoff function
        :param sensitivity_module: class or callable that builds sensitivity functions, should return nn.Module
        """
        super().__init__()

        self.n_dist = n_dist
        self.nf_in = nf_in
        self.nf_out = nf_out

        # Sensitivity module
        self.sensitivity = sensitivity_module(n_dist, mind_soft, maxd_soft, hard_cutoff)

        # Interaction weights
        self.int_weights = torch.nn.Parameter(torch.Tensor(n_dist, nf_out, nf_in))
        torch.nn.init.xavier_normal_(self.int_weights.data)

        # Self-term and bias
        self.selfint = torch.nn.Linear(nf_in, nf_out)  # includes bias term and self-interactions
        torch.nn.init.xavier_normal_(self.selfint.weight.data)

    def regularization_params(self):
        return [self.int_weights, self.selfint.weight]

    def forward(self, in_features, pair_first, pair_second, dist_pairs):
        """
        Pytorch Enforced Forward function

        :param in_features:
        :param pair_first:
        :param pair_second:
        :param dist_pairs:
        :return: Interaction output features
        """

        # Z' = (VSZ) + (WZ) + b
        n_atoms_real = in_features.shape[0]
        sense_vals = self.sensitivity(dist_pairs)
        # For HIPNN equation, the interaction term is VSZ, which we evaluate as V(E) where E=(SZ)
        # V: interaction weights
        # S: sensitivities
        # Z: input features
        # E: environment features (S*Z)

        # E = (SZ)
        env_features = custom_kernels.envsum(sense_vals, in_features, pair_first, pair_second)

        # (VSZ)
        env_features = torch.reshape(env_features, (n_atoms_real, self.n_dist * self.nf_in))
        # The weight permutation can be completely eliminated by reshaping the initialization
        weights_rs = torch.reshape(self.int_weights.permute(0, 2, 1), (self.n_dist * self.nf_in, self.nf_out))
        # Multiply the environment of each atom by weights
        features_out = torch.mm(env_features, weights_rs)

        # WZ + B
        features_out_selfpart = self.selfint(in_features)

        # VSZ + WZ + B
        features_out_total = features_out + features_out_selfpart

        return features_out_total


class InteractLayerVec(InteractLayer):
    def __init__(self, nf_in, nf_out, n_dist, mind_soft, maxd_soft, hard_cutoff, sensitivity_module):
        super().__init__(nf_in, nf_out, n_dist, mind_soft, maxd_soft, hard_cutoff, sensitivity_module)

        self.vecscales = torch.nn.Parameter(torch.Tensor(nf_out))
        torch.nn.init.normal_(self.vecscales.data)

    def forward(self, in_features, pair_first, pair_second, dist_pairs, coord_pairs):

        n_atoms_real = in_features.shape[0]
        sense_vals = self.sensitivity(dist_pairs)

        # Scalar part
        env_features = custom_kernels.envsum(sense_vals, in_features, pair_first, pair_second)
        env_features = torch.reshape(env_features, (n_atoms_real, self.n_dist * self.nf_in))
        weights_rs = torch.reshape(self.int_weights.permute(0, 2, 1), (self.n_dist * self.nf_in, self.nf_out))
        features_out = torch.mm(env_features, weights_rs)

        # Vector part
        sense_vec = sense_vals.unsqueeze(1) * (coord_pairs / dist_pairs.unsqueeze(1)).unsqueeze(2)
        sense_vec = sense_vec.reshape(-1, self.n_dist * 3)
        env_features_vec = custom_kernels.envsum(sense_vec, in_features, pair_first, pair_second)
        env_features_vec = env_features_vec.reshape(n_atoms_real * 3, self.n_dist * self.nf_in)
        features_out_vec = torch.mm(env_features_vec, weights_rs)
        features_out_vec = features_out_vec.reshape(n_atoms_real, 3, self.nf_out)
        features_out_vec = torch.square(features_out_vec).sum(dim=1) + 1e-30
        features_out_vec = torch.sqrt(features_out_vec)
        features_out_vec = features_out_vec * self.vecscales.unsqueeze(0)

        # Self interaction
        features_out_selfpart = self.selfint(in_features)

        features_out_total = features_out + features_out_vec + features_out_selfpart

        return features_out_total


class InteractLayerQuad(InteractLayerVec):
    def __init__(self, nf_in, nf_out, n_dist, mind_soft, maxd_soft, hard_cutoff, sensitivity_module):
        super().__init__(nf_in, nf_out, n_dist, mind_soft, maxd_soft, hard_cutoff, sensitivity_module)

        self.quadscales = torch.nn.Parameter(torch.Tensor(nf_out))
        torch.nn.init.normal_(self.quadscales.data)
        # upper indices of flattened 3x3 array minus the (3,3) component
        # which is not needed for a traceless tensor
        upper_ind = torch.as_tensor([0, 1, 2, 4, 5], dtype=torch.int64)
        self.register_buffer("upper_ind", upper_ind, persistent=False)  # Static, not part of module state

    def forward(self, in_features, pair_first, pair_second, dist_pairs, coord_pairs):

        n_atoms_real = in_features.shape[0]
        sense_vals = self.sensitivity(dist_pairs)

        # Scalar part
        env_features = custom_kernels.envsum(sense_vals, in_features, pair_first, pair_second)
        env_features = torch.reshape(env_features, (n_atoms_real, self.n_dist * self.nf_in))
        weights_rs = torch.reshape(self.int_weights.permute(0, 2, 1), (self.n_dist * self.nf_in, self.nf_out))
        features_out = torch.mm(env_features, weights_rs)

        # Vector part
        # Sensitivity
        rhats = coord_pairs / dist_pairs.unsqueeze(1)
        sense_vec = sense_vals.unsqueeze(1) * rhats.unsqueeze(2)
        sense_vec = sense_vec.reshape(-1, self.n_dist * 3)
        # Weights
        env_features_vec = custom_kernels.envsum(sense_vec, in_features, pair_first, pair_second)
        env_features_vec = env_features_vec.reshape(n_atoms_real * 3, self.n_dist * self.nf_in)
        features_out_vec = torch.mm(env_features_vec, weights_rs)
        # Norm and scale
        features_out_vec = features_out_vec.reshape(n_atoms_real, 3, self.nf_out)
        features_out_vec = torch.square(features_out_vec).sum(dim=1) + 1e-30
        features_out_vec = torch.sqrt(features_out_vec)
        features_out_vec = features_out_vec * self.vecscales.unsqueeze(0)

        # Quadrupole part
        # Sensitivity
        rhatsquad = rhats.unsqueeze(1) * rhats.unsqueeze(2)
        rhatsquad = (rhatsquad + rhatsquad.transpose(1, 2)) / 2
        tr = torch.diagonal(rhatsquad, dim1=1, dim2=2).sum(dim=1) / 3.0  # Add divide by 3 early to save flops
        tr = tr.unsqueeze(1).unsqueeze(2) * torch.eye(3, dtype=tr.dtype, device=tr.device).unsqueeze(0)
        rhatsquad = rhatsquad - tr
        rhatsqflat = rhatsquad.reshape(-1, 9)[:, self.upper_ind]  # Upper-diagonal part
        sense_quad = sense_vals.unsqueeze(1) * rhatsqflat.unsqueeze(2)
        sense_quad = sense_quad.reshape(-1, self.n_dist * 5)
        # Weights
        env_features_quad = custom_kernels.envsum(sense_quad, in_features, pair_first, pair_second)
        env_features_quad = env_features_quad.reshape(n_atoms_real * 5, self.n_dist * self.nf_in)
        features_out_quad = torch.mm(env_features_quad, weights_rs)  ##sum v b
        features_out_quad = features_out_quad.reshape(n_atoms_real, 5, self.nf_out)
        # Norm. (of traceless two-tensor from 5 component representation)
        quadfirst = torch.square(features_out_quad).sum(dim=1)
        quadsecond = features_out_quad[:, 0, :] * features_out_quad[:, 3, :]
        features_out_quad = 2 * (quadfirst + quadsecond)
        features_out_quad = torch.sqrt(features_out_quad + 1e-30)
        # Scales
        features_out_quad = features_out_quad * self.quadscales.unsqueeze(0)

        features_out_selfpart = self.selfint(in_features)

        features_out_total = features_out + features_out_vec + features_out_quad + features_out_selfpart

        return features_out_total



"""
System-by-system pair finders
"""

from itertools import product
import numpy as np
from scipy.spatial import KDTree
import torch

from .open import PairMemory

def wrap_points_np(coords, cell, inv_cell):
    # cell is (basis,cartesian)
    # inv is (cartesian,basis)
    projections = coords @ inv_cell
    wrapped_offset, fractional_coords = np.divmod(projections, 1)
    # wrapped_coords is (atoms,cartesian)
    wrapped_coords = fractional_coords @ cell
    return wrapped_coords, wrapped_offset.astype(np.int64)


def neighbor_list_np(cutoff, coords, cell):
    # cell is (basis,cartesian)
    # inv is (cartesian,basis)
    inv_cell = np.linalg.inv(cell)
    coords, wrapped_offset = wrap_points_np(coords, cell, inv_cell)

    eps = 1e-5
    inv_lengths = np.linalg.norm(inv_cell, axis=0)
    n_bounds = (np.floor(cutoff * inv_lengths + eps) + 1).astype(int)
    n1, n2, n3 = n_bounds

    drij = coords[:, np.newaxis] - coords[np.newaxis, :]
    wrap_offset_ij = wrapped_offset[:, np.newaxis] - wrapped_offset[np.newaxis, :]

    pair_first = []
    pair_second = []
    pair_image = []
    for i1 in range(-n1, n1 + 1):
        for i2 in range(-n2, n2 + 1):
            for i3 in range(-n3, n3 + 1):
                oi = (i1, i2, i3)
                oi_float = np.asarray(oi, dtype=coords.dtype)
                offset = np.dot(oi_float, cell)
                diff = drij + offset
                dist = np.linalg.norm(diff, axis=2)
                pf, ps = np.nonzero(dist < cutoff)

                if i1 == i2 == i3 == 0:
                    nonself = pf != ps
                    pf = pf[nonself]
                    ps = ps[nonself]

                if len(pf) > 0:
                    pi = np.repeat(np.asarray(oi)[np.newaxis, :], repeats=len(pf), axis=0)
                    # pi is wrapped image locations, need to convert back to absolute.
                    pi = pi - wrap_offset_ij[pf, ps]
                    pair_first.append(pf)
                    pair_second.append(ps)
                    pair_image.append(pi)

    pair_first = np.concatenate(pair_first)
    pair_second = np.concatenate(pair_second)
    pair_image = np.concatenate(pair_image)

    return pair_first, pair_second, pair_image


@torch.jit.script
def wrap_points_torch(coords, cell, inv_cell):
    # cell is (basis,cartesian)
    # inv is (cartesian,basis)
    projections = coords @ inv_cell
    fractional_coords = torch.remainder(projections, 1)
    wrapped_offset = torch.div(projections, 1, rounding_mode="floor")
    # wrapped_coords is (atoms,cartesian)
    wrapped_coords = fractional_coords @ cell
    return wrapped_coords, wrapped_offset.to(torch.int64)


@torch.jit.script
def compute_offset_range(inv_cell, cutoff: float):
    dev = inv_cell.device
    eps = 1e-5
    inv_lengths = torch.linalg.norm(inv_cell, dim=0)
    n_bounds = (torch.floor(cutoff * inv_lengths + eps) + 1).to(torch.int64)
    n1, n2, n3 = n_bounds.unbind()
    n1range = torch.arange(-n1, n1 + 1, device=dev)
    n2range = torch.arange(-n2, n2 + 1, device=dev)
    n3range = torch.arange(-n3, n3 + 1, device=dev)

    offset_range = torch.cartesian_prod(n1range, n2range, n3range)
    # shape: n_permutations, 3 (basis)
    return offset_range


@torch.jit.script
def neighbor_list_torch(cutoff: float, coords, cell):
    # cell is (basis,cartesian)
    # inv is (cartesian,basis)
    inv_cell = torch.linalg.inv(cell)
    coords, wrapped_offset = wrap_points_torch(coords, cell, inv_cell)

    offset_range = compute_offset_range(inv_cell, cutoff)
    # shape: n_permutations, 3 (basis)

    drij = torch.sub(coords.unsqueeze(1), coords.unsqueeze(0))
    # shape n_atoms, n_atoms, 3
    wrap_offset_ij = torch.sub(wrapped_offset.unsqueeze(1), wrapped_offset.unsqueeze(0))

    perm_offsets = offset_range.to(cell.dtype) @ cell
    # shape n_perms, 3 (cartesian)
    perm_offsets = perm_offsets.unsqueeze(1).unsqueeze(2)
    # shape n_perms, 1, 1, 3

    drij = drij.unsqueeze(0)
    # shape 1, n_atoms, n_atoms, 3

    diff = torch.add(drij, perm_offsets)

    dist = torch.linalg.norm(diff, dim=3)
    nonz = torch.nonzero(dist < cutoff)
    po, pf, ps = nonz.unbind(1)
    pi = offset_range[po]

    # Remove connections to self.
    other_image = pi.type(torch.bool).any(dim=1)  # same as (pi != 0).any(dim=1)
    other_particle = torch.ne(pf, ps)  # same as pf != ps
    nonself = torch.logical_or(other_particle, other_image)

    pf = pf[nonself]
    ps = ps[nonself]
    pi = pi[nonself]
    # pi is wrapped image locations, need to convert back to absolute.
    pi = pi - wrap_offset_ij[pf, ps]
    return pf, ps, pi

def neighbor_list_kdtree(cutoff, coords, cell):
    '''
    Use KD Tree implementation from scipy.spatial to find pairs under periodic boundary conditions 
    with an orthonormal cell.
    '''
    
    # Verify that cell is orthorhombic
    cell_prod = cell @ cell.T
    if torch.count_nonzero(cell_prod - torch.diag(torch.diag(cell_prod))):
        raise ValueError("KD Tree search only works for orthorhombic cells.")
    
    # Verify that the cutoff is less than the side lengths of the cell
    cell_side_lengths = torch.sqrt(torch.diag(cell_prod))
    if (cutoff >= cell_side_lengths).any():
        raise ValueError(f"Cutoff value ({cutoff}) must be less than the cell slide lengths ({cell_side_lengths}).")
    
    if torch.count_nonzero(cell - torch.diag(torch.diag(cell))):
        # Transform via isometry to a basis where cell is a diagonal matrix if it currently is not
        new_cell = torch.sqrt(cell_prod)
        new_coords = coords @ torch.linalg.inv(cell) @ new_cell
    else:
        new_cell = cell.clone()
        new_coords = coords.clone()

    new_coords = new_coords % torch.diag(new_cell)

    # The following three lines are included to prevent an extremely rare but not unseen edge 
    # case where the modulo operation returns a particle coordinate that is exactly equal to 
    # the corresponding cell length, causing KDTree to throw an error
    n_particles = new_coords.shape[0]
    tiled_cell = torch.tile(torch.diag(new_cell), (n_particles,)).reshape(n_particles, -1)
    new_coords = torch.where(new_coords == tiled_cell, 0, new_coords)

    new_coords = new_coords.detach().cpu().numpy()
    new_cell = torch.diag(new_cell).detach().cpu().numpy()

    tree = KDTree(
        data=new_coords, 
        boxsize=new_cell,
    )
    
    pairs = tree.query_pairs(r=cutoff, output_type='ndarray')
    pairs = torch.as_tensor(pairs, device=coords.device)
    pair_first, pair_second = torch.unbind(pairs, dim=1)

    # Find difference vector between pairs without considering the MIC
    pair_diff = torch.sub(coords[pair_first], coords[pair_second])

    # Possible adjacent offset directions for images of the difference vector
    offset_range = torch.tensor(list(product([-1, 0, 1], repeat=3)), device=coords.device)

    # All adjacent offsets
    perm_offsets = offset_range.to(cell.dtype) @ cell
    
    # All adjacent offset images of the difference vector
    pair_diff = pair_diff.unsqueeze(1) + perm_offsets.unsqueeze(0)
    
    # L2 norm of offset images
    pair_diff = torch.linalg.norm(pair_diff, dim=2)

    # Index of shortest offset image
    pair_diff = torch.argmin(pair_diff, dim=1)

    # Offset direction corresponding to shortest offset image
    pair_image = offset_range[pair_diff]

    # KDTree only returns each pair once (eg. (1,2) but not (2,1))
    doubled_pair_first = torch.concat((pair_first, pair_second))
    doubled_pair_second = torch.concat((pair_second, pair_first))
    doubled_pair_image = torch.concat((pair_image, -pair_image))

    return doubled_pair_first, doubled_pair_second, doubled_pair_image


class _DispatchNeighbors(torch.nn.Module):
    def __init__(self, dist_hard_max):
        super().__init__()
        self.dist_hard_max = dist_hard_max

        self.set_combinator(1)
        self.n_images = 1

    def set_combinator(self, n_images):
        self.n_images = n_images

    def compute_one(self, r, c):
        return NotImplemented

    def forward(self, coordinates, nonblank, real_atoms, inv_real_atoms, cell, mol_index, n_molecules, n_atoms_max):

        with torch.no_grad():
            dev = coordinates.device  # where to put the results.

            cell_list = cell.unbind(0)
            coord_list = coordinates.unbind(0)
            nb_list = nonblank.unbind(0)
            sys_inv_atoms = inv_real_atoms.reshape(n_molecules, n_atoms_max).unbind(0)
            nlist_data = []

            for (r, c, nb, sia) in zip(coord_list, cell_list, nb_list, sys_inv_atoms):
                # r: positions
                # c: cell
                # nb: nonblank
                # sia: system inv atom indices, these map from system atom numbers to batch atom numbers

                # Remove blank atoms from positions and indices of atoms.
                r = r[nb]
                sia = sia[nb]
                # dispatch to implementation
                outs = self.compute_one(r, c)
                pf, ps, of = outs

                # convert system-relative atom indices back to whole-batch atom indices
                pf = sia[pf]
                ps = sia[ps]

                # store
                nlist_data.append((pf, ps, of))

            # transpose lists into tensors for whole batch
            try:
                pair_first, pair_second, offsets = zip(*nlist_data)
                pair_first = torch.cat(pair_first).to(dev)
                pair_second = torch.cat(pair_second).to(dev)
                offsets = torch.cat(offsets).to(dev)
                max_images = offsets.abs().max()
                self.set_combinator(max_images)

            # Catch when no neighbors in this whole batch...  very rare?
            except (ValueError, RuntimeError):
                if pair_first.shape[0] != 0:
                    raise
                pair_first = torch.empty(0, dtype=torch.int64, device=dev)
                pair_second = torch.empty(0, dtype=torch.int64, device=dev)
                offsets = torch.empty((0, 3), dtype=torch.int64, device=dev)

            # Number the offsets
            n_off = self.n_images * 2 + 1
            o1, o2, o3 = (offsets + self.n_images).unbind(dim=1)
            offset_index = o3 + n_off * (o2 + n_off * o1)

            pair_mol = mol_index[pair_first]
            pair_cell = cell[pair_mol]
            pair_offsets = torch.bmm(offsets.unsqueeze(1).to(pair_cell.dtype), pair_cell).squeeze(1)

        # now calculate pair_dist, paircoord differentiably
        # print("Pairs found",pair_first.shape)
        coordflat = coordinates.reshape(n_molecules * n_atoms_max, 3)[real_atoms]
        paircoord = coordflat[pair_first] - coordflat[pair_second] + pair_offsets
        distflat2 = paircoord.norm(dim=1)

        return distflat2, pair_first, pair_second, paircoord, offsets, offset_index


class NPNeighbors(_DispatchNeighbors):
    def compute_one(self, positions, cell):
        positions = positions.detach().cpu().numpy()
        cell = cell.detach().cpu().numpy()
        outputs = neighbor_list_np(self.dist_hard_max, positions, cell)
        return [torch.from_numpy(o) for o in outputs]


class TorchNeighbors(_DispatchNeighbors):
    def compute_one(self, positions, cell):
        with torch.no_grad():
            outputs = neighbor_list_torch(self.dist_hard_max, positions, cell)
        return outputs
    
class KDTreeNeighbors(_DispatchNeighbors):
    '''
    Node for finding pairs under periodic boundary conditions using Scipy's KD Tree algorithm. 
    Cell must be orthorhombic.
    '''

    def compute_one(self, positions, cell):
        with torch.no_grad():
            outputs = neighbor_list_kdtree(self.dist_hard_max, positions, cell)
        return outputs
    

class KDTreePairsMemory(PairMemory):
    '''
    Implementation of KDTreePairs with an added memory component.

    Stores current pair indices in memory and reuses them to compute the pair distances if no 
    particle has moved more than skin/2 since last pair calculation. Otherwise uses the
    _pair_indexer_class to recompute the pairs.

    Increasing the value of 'skin' will increase the number of pair distances computed at
    each step, but decrease the number of times new pairs must be computed. Skin should be 
    set to zero while training for fastest results.
    '''

    _pair_indexer_class = KDTreeNeighbors

    def forward(self, coordinates, nonblank, real_atoms, inv_real_atoms, cells, mol_index, n_molecules, n_atoms_max):
        if self.recalculation_needed(coordinates, cells):
            self.recalculations += 1

            inputs = (coordinates, nonblank, real_atoms, inv_real_atoms, cells, mol_index, n_molecules, n_atoms_max)
            outputs = self._pair_indexer(*inputs)
            distflat2, pair_first, pair_second, paircoord, offsets, offset_index = outputs

            with torch.no_grad():
                pair_mol = mol_index[pair_first]
                pair_cell = cells[pair_mol]
                pair_offsets = torch.bmm(offsets.unsqueeze(1).to(pair_cell.dtype), pair_cell).squeeze(1)

            for name, var in [
                ("pair_first", pair_first),
                ("pair_second", pair_second),
                ("offsets", offsets),
                ("offset_index", offset_index),
                ("pair_offsets", pair_offsets),
                ("positions", coordinates),
                ("cells", cells),
                ]:
                self.__setattr__(name, var)

        else:
            self.reuses += 1

            coordflat = coordinates.reshape(n_molecules * n_atoms_max, 3)[real_atoms]
            paircoord = coordflat[self.pair_first] - coordflat[self.pair_second] + self.pair_offsets
            distflat2 = paircoord.norm(dim=1)

        return distflat2, self.pair_first, self.pair_second, paircoord, self.offsets, self.offset_index




"""
Custom Kernels for hip-nn interaction sum.

This module provides implementations in pytorch, numba, and cupy.

Pytorch implementations take extra memory, but launch faster than numba kernels.

Numba kernels use far less memory, but do come with some launching overhead on GPUs.

Cupy kernels only work on the GPU, but are faster than numba.
Cupy kernels require numba for CPU operations.
"""
import warnings
from typing import Union

from .. import settings
from . import autograd_wrapper, env_pytorch

CUSTOM_KERNELS_AVAILABLE = []
try:
    import numba

    CUSTOM_KERNELS_AVAILABLE.append("numba")
except ImportError:
    pass

try:
    import cupy

    if "numba" not in CUSTOM_KERNELS_AVAILABLE:
        warnings.warn("Cupy was found, but numba was not. Cupy custom kernels not available.")
    else:
        CUSTOM_KERNELS_AVAILABLE.append("cupy")
except ImportError:
    pass

if not CUSTOM_KERNELS_AVAILABLE:
    warnings.warn("Numba or cupy not available: Custom Kernels will be disabled.")

CUSTOM_KERNELS_ACTIVE = False

envsum, sensesum, featsum = None, None, None


def _check_numba():
    import numba.cuda
    import torch

    if not numba.cuda.is_available():
        if torch.cuda.is_available():
            warnings.warn("numba.cuda.is_available() returned False: Custom kernels will fail on GPU tensors.")
        return True
    else:
        # atexit.register(numba.cuda.close)
        # Dev note for the future: Do not attempt the above `atexit` call!
        # Causes segfault on program exit on some systems.
        # Probably due to both numba and torch trying to finalize the GPU.
        # Leaving this note here in case anyone is tempted to try it in the future.
        # (At one point this was the right strategy...)
        return False


def _check_cupy():
    import cupy
    import numba
    import torch

    if not cupy.cuda.is_available():
        if torch.cuda.is_available():
            warnings.warn("cupy.cuda.is_available() returned False: Custom kernels will fail on GPU tensors.")


def set_custom_kernels(active: Union[bool, str] = True):
    """
    Activate or deactivate custom kernels for interaction.

    :param active: If true, set custom kernels to the best available. If False, turn them off and default to pytorch.
       If "numba" or "cupy", use those implementations explicitly. If "auto", use best available.
    :return: None
    """
    global envsum, sensesum, featsum, CUSTOM_KERNELS_ACTIVE

    if isinstance(active, str):
        active = active.lower()

    if active not in [True, False, "numba", "cupy", "pytorch", "auto"]:
        raise ValueError(f"Unrecognized custom kernel implementation: {active}")

    active_map = {"auto": True, "pytorch": False}
    if not CUSTOM_KERNELS_AVAILABLE:
        if active == "auto" or active == "pytorch":
            active = False
        elif active:
            raise RuntimeError("Numba or cupy was not found. Custom kernels are not available.")
    else:
        active = active_map.get(active, active)

    if not active:
        envsum = env_pytorch.envsum
        sensesum = env_pytorch.sensesum
        featsum = env_pytorch.featsum
        CUSTOM_KERNELS_ACTIVE = False
        return

    if not CUSTOM_KERNELS_AVAILABLE:
        raise RuntimeError("Numba was not found. Custom kernels are not available.")

    if active is True:
        if "cupy" in CUSTOM_KERNELS_AVAILABLE:
            active = "cupy"
        else:
            active = "numba"

    if active not in CUSTOM_KERNELS_AVAILABLE:
        raise RuntimeError(f"Unavailable custom kernel implementation: {active}")

    if active == "cupy":
        _check_numba()
        _check_cupy()
        from .env_cupy import cupy_envsum, cupy_featsum, cupy_sensesum

        envsum, sensesum, featsum = autograd_wrapper.wrap_envops(cupy_envsum, cupy_sensesum, cupy_featsum)
    elif active == "numba":
        _check_numba()
        from .env_numba import new_envsum, new_featsum, new_sensesum

        envsum, sensesum, featsum = autograd_wrapper.wrap_envops(new_envsum, new_sensesum, new_featsum)

    else:
        # We shouldn't get here except possibly mid-development, but just in case:
        # if you add a custom kernel implementation remember to add to this
        # dispatch block.
        raise ValueError(f"Unknown Implementation: {active}")

    CUSTOM_KERNELS_ACTIVE = active


try_custom_kernels = settings.USE_CUSTOM_KERNELS
set_custom_kernels(try_custom_kernels)
del try_custom_kernels



"""
Numba implementation of envsum operations.
"""
import numba
import numba.cuda
import numpy as np

from .utils import resort_pairs_cached
from .tensor_wrapper import via_numpy, NumbaCompatibleTensorFunction

# Very basic implementation.
# While simple, it beats a set of pytorch operations simply by using far less memory.

# conventions:
# pidx  : index of pair
# pfidx : index of first atom in pair (receiver)
# psidx : index of second atom in pair (sender)
# fidx  : index of feature
# nidx  : index of sensitivity (nu)


# Kernel which sums sensitivities and features to get environment.
# Numpy core signature: (p,n),(a,f),(p),(p),(a,n,f)
class WrappedEnvsum(NumbaCompatibleTensorFunction):
    def __call__(self, *args, **kwargs):
        sense, feat, pfirst, psecond = args
        psecond_hold = psecond
        argsort, atom1_ids, atom1_starts, pfirst, (sense, psecond) = resort_pairs_cached(pfirst, [sense, psecond])
        resort_pairs_cached(psecond_hold, [])
        args = sense, feat, pfirst, psecond, atom1_ids, atom1_starts
        return super().__call__(*args, **kwargs)

    def out_shape(self, sense_shape, feat_shape, *other_shapes):
        n_pair, n_nu = sense_shape
        n_atom, n_feature = feat_shape
        return n_atom, n_nu, n_feature

    def launch_bounds(self, sense_shape, fs, pfs, pss, atom1_ids_shape, *other_shapes):
        n_pairs, n_nu = sense_shape
        n_atom, n_feat = fs
        (n_atoms_interacting,) = atom1_ids_shape
        TPB_MAX = 512
        TPB_X = n_feat
        TPB_Y = TPB_MAX // n_feat
        TPB = (TPB_X, TPB_Y)
        BPG_X = (n_atoms_interacting + TPB_Y - 1) // TPB_Y
        BPG_Y = 1
        BPG = (BPG_X, BPG_Y)
        return BPG, TPB

    @staticmethod
    def make_kernel(KERNEL_DTYPE):
        sig = "void({DTYPE}[:,:,],{DTYPE}[:,:],int64[:],int64[:],int64[:],int64[:],{DTYPE}[:,:,:])".format(
            DTYPE=KERNEL_DTYPE
        )

        @numba.cuda.jit(
            sig,
        )
        def kernel(sens, feat, pfirst, psecond, atom1_ids, atom1_starts, env):
            n_pairs, n_nu = sens.shape
            n_atom, n_feat = feat.shape
            (n_atoms_interacting,) = atom1_ids.shape

            tx = numba.cuda.threadIdx.x
            sx = numba.cuda.blockDim.x
            by = numba.cuda.blockIdx.y
            fidx = by * sx + tx

            ty = numba.cuda.threadIdx.y
            sy = numba.cuda.blockDim.y
            bx = numba.cuda.blockIdx.x
            i = bx * sy + ty
            if i < n_atoms_interacting:
                pair_start = atom1_starts[i]
                pair_end = atom1_starts[i + 1]
                pfidx = atom1_ids[i]
                for nidx in range(n_nu):
                    out = KERNEL_DTYPE(0.0)
                    for pidx in range(pair_start, pair_end):
                        psidx = psecond[pidx]
                        out += sens[pidx, nidx] * feat[psidx, fidx]
                    env[pfidx, nidx, fidx] = out

        return kernel

    @staticmethod
    @via_numpy
    @numba.jit(nopython=True, parallel=True)
    def cpu_kernel(sens, feat, pfirst, psecond, atom_ids, atom_starts):

        n_pairs, n_nu = sens.shape
        n_atom, n_feat = feat.shape
        (n_atom_with_pairs,) = atom_ids.shape

        env_features = np.zeros((n_atom, n_nu, n_feat), dtype=sens.dtype)
        for i in numba.prange(n_atom_with_pairs):
            pfstart = atom_starts[i]
            pfend = atom_starts[i + 1]
            pfidx = atom_ids[i]
            tmp = np.zeros((n_nu, n_feat), dtype=sens.dtype)
            for pidx in range(pfstart, pfend):
                psidx = psecond[pidx]
                for nidx in numba.prange(n_nu):
                    s = sens[pidx, nidx]
                    if abs(s) > 1e-10:
                        for fidx in range(n_feat):
                            tmp[nidx, fidx] += s * feat[psidx, fidx]
            env_features[pfidx] = tmp
        return env_features


# Kernel which sums environment and features to get sensitivity
# Numpy core signature: (a,n,f),(a,f),(p),(p),(p,n),
class WrappedSensesum(NumbaCompatibleTensorFunction):
    def out_shape(self, env_shape, feat_shape, pfirst_shape, psecond_shape):
        (n_pair,) = pfirst_shape
        n_atom, n_nu, n_feat = env_shape
        return n_pair, n_nu

    def launch_bounds(self, env_shape, feat_shape, pfirst_shape, psecond_shape):
        (n_pairs,) = pfirst_shape
        n_atoms, n_nu, n_feat = env_shape
        TPB_MAX = 512
        TPB_Y = n_nu
        TPB_X = TPB_MAX//TPB_Y
        TPB = (TPB_X,TPB_Y)
        BPG = (n_pairs + TPB_X -1 )//TPB_X
        return BPG, TPB

    @staticmethod
    def make_kernel(KERNEL_DTYPE):
        sig = "void({DTYPE}[:,:,:],{DTYPE}[:,:],int64[:], int64[:],{DTYPE}[:,:])".format(DTYPE=KERNEL_DTYPE)

        @numba.cuda.jit(
            sig,
        )
        def kernel(env, feat, pfirst, psecond, sense):
            (n_pairs,) = pfirst.shape
            n_atom, n_nu, n_feat = env.shape
            pidx, nidx, fidx = numba.cuda.grid(3)

            if pidx < n_pairs:
                pfidx = pfirst[pidx]
                psidx = psecond[pidx]
                tmp = KERNEL_DTYPE(0.0)
                for fidx in range(n_feat):
                    tmp += env[pfidx, nidx, fidx] * feat[psidx, fidx]
                sense[pidx, nidx] = tmp

        return kernel

    @staticmethod
    @via_numpy
    @numba.jit(nopython=True, parallel=True)
    def cpu_kernel(env, feat, pfirst, psecond):
        n_atom, n_nu, n_feat = env.shape
        (n_pairs,) = pfirst.shape
        sense = np.zeros((n_pairs, n_nu), dtype=env.dtype)
        for pidx in numba.prange(n_pairs):
            pfidx = pfirst[pidx]
            psidx = psecond[pidx]
            for nidx in numba.prange(n_nu):
                tmp = 0
                for fidx in range(n_feat):
                    tmp += env[pfidx, nidx, fidx] * feat[psidx, fidx]
                sense[pidx, nidx] = tmp
        return sense


# Kernel which sums environment and sensitivity to get features
# Numpy core signature: (a,n,f),(p,n),(p),(p),(a,f),
class WrappedFeatsum(NumbaCompatibleTensorFunction):
    def __call__(self, *args, **kwargs):
        env, sense, pfirst, psecond = args
        pfirst_hold = pfirst
        argsort, atom2_ids, atom2_starts, psecond, (sense, pfirst) = resort_pairs_cached(psecond, [sense, pfirst])
        resort_pairs_cached(pfirst_hold, [])
        args = env, sense, pfirst, psecond, atom2_ids, atom2_starts
        return super().__call__(*args, **kwargs)

    def out_shape(self, env_shape, *other_shapes):
        n_atom, n_nu, n_feature = env_shape
        return n_atom, n_feature

    def launch_bounds(self, env_shape, sense_shape, pfirst_shape, psecond_shape, atom2_id_shape, atom2_startshape):
        # Note, this kernel has transposed launch bounds:
        # The blocks in the x direction go with threads in the
        # y direction, and likewise for y blocks-> x threads
        # This is done because we want features along the x direction for coherent warps
        # But we want atoms along the block x direction to avoid
        # problems when reaching too many atoms (can only launch 65k blocks in y direction...)
        n_pairs, n_nu = sense_shape
        n_atom, n_nu, n_feat = env_shape
        TPB_max = 512
        if n_feat > 32:
            TPB_x = ((n_feat + 31) // 32) * 32
            TPB_y = TPB_max // TPB_x
            BPG_y = 1
        else:
            # If there aren't enough features to really fill a warp, stride them across the blocks instead.
            TPB_x = 1
            TPB_y = TPB_max
            BPG_y = n_feat
        TPB = (TPB_x, TPB_y)
        BPG_x = (n_atom + TPB_y - 1) // TPB_y
        BPG = (BPG_x, BPG_y)
        return BPG, TPB

    @staticmethod
    def make_kernel(KERNEL_DTYPE):
        sig = "void({DTYPE}[:,:,:],{DTYPE}[:,:],int64[:],int64[:],int64[:],int64[:],{DTYPE}[:,:])".format(
            DTYPE=KERNEL_DTYPE
        )

        @numba.cuda.jit(
            sig,
        )
        def kernel(env, sense, pfirst, psecond, atom2_ids, atom2_starts, feat):
            n_pairs, n_nu = sense.shape
            n_atom, n_feat = feat.shape
            (n_atoms_with_pairs,) = atom2_ids.shape

            tx = numba.cuda.threadIdx.x
            sx = numba.cuda.blockDim.x
            by = numba.cuda.blockIdx.y
            fidx = by * sx + tx

            ty = numba.cuda.threadIdx.y
            sy = numba.cuda.blockDim.y
            bx = numba.cuda.blockIdx.x
            aidx = bx * sy + ty

            if aidx < n_atoms_with_pairs and fidx < n_feat:
                pair_start = atom2_starts[aidx]
                pair_end = atom2_starts[aidx + 1]
                psidx = atom2_ids[aidx]
                tmp = KERNEL_DTYPE(0.0)
                for pidx in range(pair_start, pair_end):
                    pfidx = pfirst[pidx]
                    for nidx in range(n_nu):
                        # env call coalesces across the warp, sense call broadcasts across warp
                        tmp += env[pfidx, nidx, fidx] * sense[pidx, nidx]
                feat[psidx, fidx] = tmp

        return kernel

    @staticmethod
    @via_numpy
    @numba.jit(nopython=True, parallel=True)
    def cpu_kernel(
        env,
        sens,
        pfirst,
        psecond,
        atom_ids,
        atom_starts,
    ):
        n_atom, n_nu, n_feat = env.shape
        (n_pairs,) = pfirst.shape
        (n_atoms_with_pairs,) = atom_ids.shape
        feat = np.zeros((n_atom, n_feat), dtype=sens.dtype)

        for i in numba.prange(n_atoms_with_pairs):
            psstart = atom_starts[i]
            psend = atom_starts[i + 1]
            psidx = atom_ids[i]
            tmp = np.zeros(n_feat, dtype=sens.dtype)
            for pidx in range(psstart, psend):
                pfidx = pfirst[pidx]
                for nidx in range(n_nu):
                    for fidx in numba.prange(n_feat):
                        tmp[fidx] += env[pfidx, nidx, fidx] * sens[pidx, nidx]
            feat[psidx] = tmp
        return feat


new_envsum = WrappedEnvsum()
new_sensesum = WrappedSensesum()
new_featsum = WrappedFeatsum()



"""
Pure pytorch implementation of envsum operations
"""
import torch


def envsum(sensitivities, features, pair_first, pair_second):
    n_pairs, n_nu = sensitivities.shape
    n_atom, n_feat = features.shape
    pair_features = features[pair_second].unsqueeze(1)
    sensitivities = sensitivities.unsqueeze(2)
    pair_env_features = sensitivities * pair_features
    env_features = torch.zeros((n_atom, n_nu, n_feat), device=features.device, dtype=features.dtype)
    env_features.index_add_(0, pair_first, pair_env_features)
    return env_features


def sensesum(env, features, pair_first, pair_second):
    n_atoms, n_nu, n_feat = env.shape
    pair_env = env[pair_first]
    pair_feat = features[pair_second]
    sense = (pair_env * pair_feat.unsqueeze(1)).sum(dim=2)
    return sense


def featsum(env, sense, pair_first, pair_second):
    n_atoms, n_nu, n_feat = env.shape
    pair_env = env[pair_first]
    pair_feat = (pair_env * sense.unsqueeze(2)).sum(dim=(1,))
    feat = torch.zeros(n_atoms, n_feat, device=env.device, dtype=env.dtype)
    feat.index_add_(0, pair_second, pair_feat)
    return feat



"""
Nodes for inputting information to the graph.
"""
import warnings
from .base import InputNode
from ..indextypes import IdxType
from .tags import Charges, Positions, Species, PairCache


class SpeciesNode(Species, InputNode):
    _index_state = IdxType.MolAtom
    input_type_str = "Species"


class PositionsNode(Positions, InputNode):
    _index_state = IdxType.MolAtom
    input_type_str = "Positions"


class CellNode(InputNode):
    _index_state = IdxType.Molecules
    input_type_str = "Cells"

class ForceNode(InputNode):
    _index_state = IdxType.MolAtom
    input_type_str = "Force"


class InputCharges(Charges, InputNode):
    _index_state = IdxType.MolAtom
    input_type_str = "InputCharges"


class Indices(InputNode):
    _index_state = IdxType.Molecules
    input_type_str = "Index"

    def __init__(self):
        super().__init__(db_name="indices")


class PairIndices(PairCache, InputNode):
    _index_state = IdxType.NotFound
    input_type_str = "PairIndices"


class SplitIndices(InputNode):
    _index_state = IdxType.Molecules
    input_type_str = "Index"

    def __init__(self):
        super().__init__(db_name="indices")



"""
Graph Operations ("gops") that process or transform a set of nodes.
"""
import copy

from .nodes.base import InputNode, MultiNode
from .nodes.base.algebra import ValueNode
from .nodes.base.node_functions import NodeNotFound, NodeOperationError
from .indextypes import soft_index_type_coercion
from . import get_connected_nodes, find_unique_relative


def get_subgraph(required_nodes):
    """
    Get the subgraph associated with some target (required) nodes.

    :param required_nodes: the nodes to compute

    :return: a list of nodes involved in the computation of the required nodes.
    """

    required_nodes = list(set(required_nodes))
    subgraph_nodes = required_nodes + [p for node in required_nodes for p in node.get_all_parents()]

    subgraph_nodes = subgraph_nodes + [c for mn in subgraph_nodes if isinstance(mn, MultiNode) for c in mn.children]
    # ^-- a note on this:
    # Children are included because bad things happen if a multinode is left without any of its children;
    # The API using dot syntax will break.
    # Adding these to a GraphModule is not expensive - it adds a bit of unwrapping time.

    return list(set(subgraph_nodes))


def compute_evaluation_order(all_nodes):
    """
    Computes the evaluation order for forward computation.

    :param all_nodes: The nodes to compute

    :return:  outputs_list, inputs_list
        Outputs list is the node to evaluate next, which is fed the inputs in the
        corresponding entry of inputs_list

    """

    # The strategy here is to emulate trying everything, and record an order that works.
    # It's computationally inefficient, but it's very easy to implement.
    # The strategy is greedy and brute-force --  it won't scale to large numbers of nodes."""

    evaluation_inputs_list = []
    evaluation_outputs_list = []
    unsatisfied_nodes = all_nodes.copy()
    satisfied_nodes = set()
    n = -1
    while len(unsatisfied_nodes) > 0:
        n += 1

        # Consider each node we haven't found yet computable
        for considered_node in unsatisfied_nodes.copy():
            parents = considered_node.parents
            uncomputed_parents = [x for x in parents if x not in satisfied_nodes]

            # If that node has zero uncomputed parents, we can add it to the evaluation order.
            if len(uncomputed_parents) == 0:
                unsatisfied_nodes.remove(considered_node)
                satisfied_nodes.add(considered_node)
                # Input nodes need not be considered; their values are given to `forward()`
                if not isinstance(considered_node, InputNode):
                    evaluation_inputs_list.append(parents)
                    evaluation_outputs_list.append(considered_node)
                break  # This break is necessary to make the `else` clause work;
                # if we find a node, we break and the `else` statement won't be hit.
        else:
            # For-else structure will land here if no satisfiable nodes are found during the iteration.
            # If we made it through all the unsatisfied nodes, something must be wrong.
            raise ValueError("Graph inputs do not seem to specify outputs! Iteration: {}".format(n))

    check_evaluation_order(evaluation_outputs_list, evaluation_inputs_list)
    return evaluation_outputs_list, evaluation_inputs_list


def check_evaluation_order(evaluation_outputs_list, evaluation_inputs_list):
    """
    Validate an evaluation order.

    :param evaluation_outputs_list:
    :param evaluation_inputs_list:
    :return:
    """
    pseudo_evaluated = list()
    for out_node, inputs_for_node in zip(evaluation_outputs_list, evaluation_inputs_list):
        for in_node in inputs_for_node:
            if in_node not in pseudo_evaluated:
                if not isinstance(in_node, InputNode):
                    raise AssertionError("Nodes are not ordered correctly.")
                else:
                    pseudo_evaluated.append(in_node)
        pseudo_evaluated.append(out_node)


def copy_subgraph(required_nodes, assume_inputed, tag=None):
    """
    Copy a subgraph for the required nodes.
    Doesn't copy the modules.
    Returns copies of the specified nodes, linked implicitly to copies of their parents.

    If assume_inputted is not empty, this will result in partially disconnected nodes.
    As such, we recommend running `check_link_consistency` after
    you have finished your graph-level operations.

    :param required_nodes: output nodes of the computation
    :param assume_inputed: nodes whose parents we will not include in the copy (these will be left dangling)
    :param tag: a string to preprend to the resulting node's names.
    :return: new_required, new_subgraph
        ``new_required``: a list containing the new copies of ``required_nodes`` in the new subgraph.

        ``new_subgraph``: a list containing new copies of all nodes in the subgraph.
    """
    subgraph_nodes = get_subgraph(required_nodes)

    # Maps old nodes to new nodes.
    node_mapping = {n: copy.copy(n) for n in subgraph_nodes}  # Shallow copy is still linked to the same torch modules
    if tag:
        for n in node_mapping.values():
            n.name = "({}){}".format(tag, n.name)

    # Set the parents and children of the new nodes to point to the new copies.
    # Children not used in the subgraph are dropped
    for n_old, n_new in node_mapping.items():
        n_new.parents = tuple(node_mapping[p] for p in n_old.parents)
        n_new.children = tuple(node_mapping[c] for c in n_old.children if c in node_mapping)
        # print()
        # print("Old node: ",n_old)
        # print("\t parents:",n_old.parents)
        # print("\t children:",n_old.children)
        # print("New node:",n_new)
        # print("\t parents:",n_new.parents)
        # print("\t children:",n_new.children)

    new_subgraph = [node_mapping[n] for n in subgraph_nodes]
    new_required = [node_mapping[r] for r in required_nodes]

    ### Checks that the operation correctly worked.

    # Graph is complete
    assert all(x in new_subgraph for x in get_connected_nodes(new_subgraph))

    # Nothing from the old graph is in the new graph
    assert all(x not in new_subgraph for x in get_connected_nodes(required_nodes))
    # None of the parents of cut-off nodes

    # For each parent p of the assumed inputted:
    # If p not a a parent of an unassumed node, then p should not be in the graph.
    assert all(
        p not in new_subgraph
        for ai in assume_inputed
        for p in ai.parents
        if all(p not in n.parents for n in new_subgraph if n not in assume_inputed)
    )

    # print(get_connected_nodes(new_required)) # Prior debug... change to logging?
    return new_required, new_subgraph


class GraphInconsistency(NodeOperationError):
    pass


def check_link_consistency(node_set):
    """
    Make sure that back-links in the graph are correct.
    Raises GraphInconsistency if the graph is not consistent.
    All connected nodes to the node set are analyzed.
    The links are consistent if each pair of child->parent has
    a corresponding link parent<-child.

    :param node_set: iterable of nodes to check

    :raises GraphInconsistency: if the graph is not consistent.

    :return: True
    """
    node_set = set(node_set)
    node_set = get_connected_nodes(node_set)
    forward_links = set((parent, child) for parent in node_set for child in parent.children)
    back_links = set((parent, child) for child in node_set for parent in child.parents)

    if forward_links == back_links:
        return True

    broken_forward = forward_links - back_links
    broken_back = back_links - forward_links
    raise GraphInconsistency(
        "Graph is inconsistent!\n" f"Broken forward links:{broken_forward}" f"Broke back links:{broken_back}"
    )


def replace_node(old_node, new_node, disconnect_old=False):
    """
    :param old_node: Node to replace
    :param new_node: Node to insert
    :param disconnect_old: If True, remove the old node connections from the graph.

    Replaces the children of old node with new node. Effectively,
    this just means going to the children and swapping out their parents to point to
    the new node. Ignores the children of old node that are parents of the new node
    in order to prevent cycles in the graph.

    If disconnect_old, remove references to the old node -- it will become unusable.

    .. Warning::
       This function changes the graph structure in-place.
       As such, if this function raises an error, it may result in a corrupted graph state.

    .. Warning::
       This function will try to coerce index states where possible. If the index types
       of the nodes are not listed, the index state will not be modified. If they are incompatible,
       this means that the graph will not function correctly.

    :return: None
    """

    new_node_requires = set(new_node.get_all_parents())

    if disconnect_old:
        if old_node in new_node_requires:
            raise NodeOperationError(
                "Cannot replace this node and remove the old one, because the"
                f"new node {new_node} depends on the old node {old_node}."
            )

    if isinstance(old_node, MultiNode):
        if not isinstance(new_node, MultiNode):
            raise TypeError("MultiNodes cannot be replaced with single nodes.")
        # Multi-nodes should always keep their children. To replace a multinode with
        # another one, we attempt to match their children with each other and perform
        # the replacement operation on the children.

        for c, cprime in _determine_multinode_child_match(old_node, new_node).items():
            replace_node(c, cprime, disconnect_old=False)

        if disconnect_old:
            old_node.disconnect()

    else:
        # If new node is a multinode, this will ensure we swap the main output in
        new_node = new_node.main_output
        # Convert index state if possible.
        new_node = soft_index_type_coercion(new_node, old_node._index_state)

        # Find children that need replacing
        swap_children = set(old_node.children) - set(new_node_requires)
        # Actually replace them
        for c in swap_children:
            c.swap_parent(old_node, new_node)

        if disconnect_old:
            old_node.disconnect()


def _determine_multinode_child_match(old_node: MultiNode, new_node: MultiNode):
    """
    Try to determine match between multinode children for `replace_node` on multinodes.
    :param old_node:
    :param new_node:
    :return:
    """

    try:
        # Try name-based matching first.
        matches = {getattr(old_node, name): getattr(new_node, name) for name in new_node._output_names}
    except AttributeError:
        # If name-based matching does not work, just match by order.
        matches = {co: cn for co, cn in zip(old_node.children, new_node.children)}
        # Raise an error if this match would leave a residual child used on the old node.

    # Raise error if there are children of the old node which do not match,
    # but are still needed for a computation.
    if any(node not in matches and len(node.children) > 0 for node in old_node.children):
        raise NodeOperationError(f"A match cannot be made between {old_node} and {new_node} children.")

    return matches


def replace_node_with_constant(node, value, name=None):
    vnode = ValueNode(value)
    vnode.name = name
    replace_node(node, vnode)


def search_by_name(nodes, name_or_dbname):
    """
    Look for a unique related node with a given db_name or name.
    The db_name will be given higher precedence.

    :param nodes: starting point for search
    :param name_or_dbname: name or dbname for node search
    :return: node that matches criterion

    Raises NodeAmbiguityError if more than one node found
    Raises NodeNotFoundError if no nodes found

    """
    try:
        return find_unique_relative(nodes, lambda n: n.db_name == name_or_dbname)
    except NodeNotFound:
        return find_unique_relative(nodes, lambda n: n.name == name_or_dbname)



"""
This file implements various stepping protocols used by different optimizer APIs.

In particular:
    - The "standard" step function which only requires that backwards has been called.
    - The "closure" step function for when line search is required (currently only active on LBFGS)
    - The "two step" style of Sharpness Aware Minimization algorithms

The main output function here is `get_step_function(optimizer)-> callable`.

The various step functions are provided as classes that act with staticmethods.
This is to provide for the possibility of extension, for example, to schemes with
stepping schemes that require additional state, or for the possibility to specifiy
the step function explicitly within the controller.
"""
from torch.optim import Optimizer, LBFGS


def standard_step_fn(optimizer, model, loss, batch_inputs, batch_targets):
    optimizer.zero_grad(set_to_none=True)
    batch_model_outputs = model(*batch_inputs)

    # The extra .mean call here deals with an edge case for multi-GPU DataParallel with scalar outputs
    batch_train_loss = loss(*batch_model_outputs, *batch_targets)[0].mean()

    batch_train_loss.backward()
    optimizer.step()
    return batch_model_outputs


def twostep_step_fn(optimizer, model, loss, batch_inputs, batch_targets):
    # Step function for SAM algorithm.
    optimizer.zero_grad(set_to_none=True)

    batch_model_outputs = model(*batch_inputs)
    batch_train_loss = loss(*batch_model_outputs, *batch_targets)[0].mean()
    batch_train_loss.backward()
    optimizer.first_step(zero_grad=True)

    batch_model_outputs_2 = model(*batch_inputs)
    loss(*batch_model_outputs_2, *batch_targets)[0].mean().backward()
    optimizer.second_step(zero_grad=True)
    return batch_model_outputs


def closure_step_fn(optimizer, model, loss, batch_inputs, batch_targets):
    return_outputs = None

    def closure():
        nonlocal return_outputs
        optimizer.zero_grad(set_to_none=True)
        batch_model_outputs = model(*batch_inputs)
        if return_outputs is None:
            return_outputs = batch_model_outputs
        batch_train_loss = loss(*batch_model_outputs, *batch_targets)[0].mean()
        batch_train_loss.backward()
        return batch_train_loss

    optimizer.step(closure)
    return return_outputs


# Note: The staticmethod version here can be re-written using class parameters
# and __init_subclass, but will they always be staticmethods?
class StepFn:
    step = NotImplemented

    def __call__(self, *args, **kwargs):
        return self.step(*args, **kwargs)


class StandardStep(StepFn):
    step = staticmethod(standard_step_fn)


class TwoStep(StepFn):
    step = staticmethod(twostep_step_fn)


class ClosureStep(StepFn):
    step = staticmethod(closure_step_fn)


def get_step_function(optimizer: Optimizer) -> callable:
    if type(optimizer).__name__ == "SAM":
        return TwoStep()
    if isinstance(optimizer, (LBFGS,)):
        return ClosureStep()
    else:
        return StandardStep()



"""
Pytorch modules that represent composable pieces of models.
"""
from . import hiplayers
from . import indexers
from . import targets
from . import transform
from . import physics
from . import excited


"""
CuPy implementation of envsum custom kernels for GPU.
"""
# Dev Note: CPU implementation of these ops is still done by numba.
# As such, numba is still required and calls to CPU ops must
# obey the same API as the numba implementations.

import torch
import cupy

from hippynn.custom_kernels.env_numba import WrappedEnvsum, WrappedSensesum, WrappedFeatsum
from hippynn.custom_kernels.utils import resort_pairs_cached

CUPY_KERNEL_CODE=r"""
extern "C" __global__
void cupy_envsum(
  const FLOATX* sense,
  const FLOATX* feat,
  const long* psecond,
  const long* atom1_ids,
  const long* atom1_starts,
  FLOATX* env,
  long n_nu,
  long n_feat,
  long n_interact
  ) {
  
  int fidx = blockIdx.y*blockDim.x + threadIdx.x;
  int i = blockIdx.x*blockDim.y + threadIdx.y;

  if (i < n_interact){
    long pair_start = atom1_starts[i];
    long pair_end = atom1_starts[i + 1];
    long pfidx = atom1_ids[i];
    for (long nidx=0; nidx<n_nu; nidx++){
      FLOATX out = 0.0;
      for (long pidx=pair_start; pidx<pair_end; pidx++){
        long psidx = psecond[pidx];
        long where_s = pidx * n_nu + nidx;
        long where_f = psidx * n_feat + fidx;
        out += sense[where_s] * feat[where_f];
      }
      long where_e = n_nu * n_feat * pfidx + n_feat * nidx + fidx;
      env[where_e] = out;
    }
  }
}

extern "C" __global__
void cupy_sensesum(
  const FLOATX* env,
  const FLOATX* feat,
  const long* pfirst,
  const long* psecond,
  FLOATX* sense,
  long n_pairs,
  long n_nu,
  long n_feat
) {
  
  long pidx = blockIdx.x*blockDim.x+threadIdx.x;
  long nidx = blockIdx.y*blockDim.y+threadIdx.y;

  if (pidx < n_pairs){
    long pfidx = pfirst[pidx];
    long psidx = psecond[pidx];
    FLOATX tmp = 0.0;
    for (long fidx=0; fidx<n_feat; fidx++){
      long where_e = n_nu*n_feat*pfidx + n_feat*nidx + fidx;
      long where_f = n_feat*psidx + fidx;
      tmp += env[where_e]*feat[where_f];
    }
    long where_s = n_nu*pidx + nidx;
    sense[where_s] = tmp;
  }
}

extern "C" __global__
void cupy_featsum(
  const FLOATX* env,
  const FLOATX* sense,
  const long* pfirst,
  const long* atom2_ids,
  const long* atom2_starts,
  FLOATX* feat,
  long n_nu,
  long n_feat,
  long n_interact
) {
  
  long fidx = blockIdx.y*blockDim.x+threadIdx.x;
  long aidx = blockIdx.x*blockDim.y+threadIdx.y;

  if (aidx < n_interact and fidx < n_feat){
    long pair_start = atom2_starts[aidx];
    long pair_end = atom2_starts[aidx + 1];
    long psidx = atom2_ids[aidx];
    FLOATX tmp = 0.0;
    for (long pidx=pair_start; pidx<pair_end; pidx++){
      long pfidx = pfirst[pidx];
      for (long nidx=0; nidx < n_nu; nidx++){
        long where_e = n_nu*n_feat*pfidx + n_feat*nidx + fidx;
        long where_s = pidx*n_nu+nidx;
        tmp += env[where_e] * sense[where_s];
      }
    }
    long where_f = psidx*n_feat + fidx;
    feat[where_f] = tmp;
  }
}
"""

_CUPY_MODULES = {
  dtype: cupy.RawModule(code=CUPY_KERNEL_CODE.replace("FLOATX",dtype))
  for dtype in ('float','double')
  }
    
def _cupy_gpu_not_found(*args,**kwargs):
    raise RuntimeError("Error: CuPy could not find the GPU."
                   "Verify that your numba installation is able to find cuda toolkit, as this \n"
                   "error condition likely indicates that torch can find the GPU, but cupy can't.\n"
                   "Alternatively, disable custom kernels.")

class CupyGPUKernel:
    _cupy_name = None
    def __init__(self):
      if not cupy.cuda.is_available():
          self.kernel32 = _cupy_gpu_not_found
          self.kernel64 = _cupy_gpu_not_found
      else:
          self.kernel32 = _CUPY_MODULES['float'].get_function(self._cupy_name)
          self.kernel64 = _CUPY_MODULES['double'].get_function(self._cupy_name)
    
    def __call__(self,dtype,BPG,TPB,array_args,shape_args):
    
        out_array = array_args[-1]
        array_args = [cupy.asarray(a.detach().contiguous()).ravel() for a in array_args]
        args = (*array_args, *shape_args)
        
        if dtype == torch.float32:
            self.kernel32(BPG,TPB,args)
        elif dtype == torch.float64:
            self.kernel64(BPG,TPB,args)
        else:
            raise ValueError("Bad dtype: {}".format(dtype))

        return out_array
      

class CupyEnvsum(CupyGPUKernel,WrappedEnvsum):
    _cupy_name = 'cupy_envsum'
    def __call__(self,sense,feat,pfirst,psecond):
        psecond_hold = psecond
        argsort, atom1_ids, atom1_starts, pfirst, (sense, psecond) = resort_pairs_cached(pfirst, [sense, psecond])
        resort_pairs_cached(psecond_hold,[])
        dev = sense.device
        if dev.type == "cpu":
            return self.cpu_kernel(sense, feat, pfirst, psecond, atom1_ids, atom1_starts)

        n_pairs, n_nu = sense.shape
        n_atoms, n_feat = feat.shape
        n_interact, = atom1_ids.shape
        dtype = sense.dtype

        env_out = torch.zeros((n_atoms,n_nu,n_feat,), device=dev, dtype=dtype)
        array_args = sense,feat,psecond,atom1_ids,atom1_starts,env_out
        shape_args = n_nu,n_feat,n_interact
        
        TPB_MAX = 512
        TPB_X = n_feat
        TPB_Y = TPB_MAX // n_feat
        TPB = (TPB_X, TPB_Y)
        BPG_X = (n_interact + TPB_Y - 1) // TPB_Y
        BPG_Y = 1
        BPG = (BPG_X, BPG_Y)
        
        args = *array_args,*shape_args
        return super().__call__(dtype,BPG,TPB,array_args,shape_args)

class CupySensesum(CupyGPUKernel,WrappedSensesum):
    _cupy_name = 'cupy_sensesum'
    def __call__(self,env,feat,pfirst,psecond):
        dev = env.device
        if dev.type == "cpu":
            return self.cpu_kernel(env, feat, pfirst, psecond)

        n_pairs, = pfirst.shape
        n_atoms, n_nu, n_feat = env.shape
        dtype = env.dtype

        sense_out = torch.zeros((n_pairs,n_nu), device=dev, dtype=dtype)
        array_args = env, feat, pfirst, psecond, sense_out
        shape_args = n_pairs, n_nu, n_feat
        
        TPB_MAX = 512
        TPB_Y = n_nu
        TPB_X = TPB_MAX//TPB_Y
        TPB = (TPB_X,TPB_Y)
        BPG_X = (n_pairs + TPB_X - 1 )//TPB_X
        BPG = (BPG_X,1)

        return super().__call__(dtype,BPG,TPB,array_args,shape_args)


class CupyFeatsum(CupyGPUKernel,WrappedFeatsum):
    _cupy_name = 'cupy_featsum'
    def __call__(self,env,sense,pfirst,psecond):
        pfirst_hold = pfirst
        argsort, atom2_ids, atom2_starts, psecond, (sense, pfirst) = resort_pairs_cached(psecond, [sense, pfirst])
        resort_pairs_cached(pfirst_hold,[])
        dev = env.device
        if dev.type == "cpu":
            return self.cpu_kernel(env, sense, pfirst, psecond, atom2_ids, atom2_starts)

        n_pairs, = pfirst.shape
        n_atoms, n_nu, n_feat = env.shape
        n_interact, = atom2_ids.shape
        dtype = env.dtype

        feat_out = torch.zeros((n_atoms,n_feat), device=dev, dtype=dtype)
        array_args = env, sense, pfirst, atom2_ids, atom2_starts, feat_out
        shape_args = n_nu, n_feat ,n_interact
        
        TPB_max = 512
        if n_feat > 32:
            TPB_x = ((n_feat + 31) // 32) * 32
            TPB_y = TPB_max // TPB_x
            BPG_y = 1
        else:
            # If there aren't enough features to really fill a warp, stride them across the blocks instead.
            TPB_x = 1
            TPB_y = TPB_max
            BPG_y = n_feat
        TPB = (TPB_x, TPB_y)
        BPG_x = (n_atoms + TPB_y - 1) // TPB_y
        BPG = (BPG_x, BPG_y)
        
        return super().__call__(dtype,BPG,TPB,array_args,shape_args)
        
cupy_envsum = CupyEnvsum()
cupy_sensesum = CupySensesum()
cupy_featsum = CupyFeatsum()



from typing import Tuple
import torch

from ...layers import excited as excited_layers
from .. import IdxType, find_unique_relative
from .base import AutoKw, SingleNode, ExpandParents, MultiNode
from .loss import _BaseCompareLoss
from .tags import Energies, HAtomRegressor, Network, AtomIndexer
from ...layers import physics as physics_layers


class NACRNode(AutoKw, SingleNode):
    """
    Compute the non-adiabatic coupling vector multiplied by the energy difference
    between two states.
    """

    _input_names = "charges i", "charges j", "coordinates", "energy i", "energy j"
    _auto_module_class = excited_layers.NACR

    def __init__(self, name: str, parents: Tuple, module="auto", module_kwargs=None, **kwargs):
        """Automatically build the node for calculating NACR * ΔE between two states i
        and j.

        :param name: name of the node
        :type name: str
        :param parents: parents of the NACR node in the sequence of (charges i, \
                charges j, positions, energy i, energy j)
        :type parents: Tuple
        :param module: _description_, defaults to "auto"
        :type module: str, optional
        :param module_kwargs: keyword arguments passed to the corresponding layer,
            defaults to None
        :type module_kwargs: dict, optional
        """

        self.module_kwargs = {}
        if module_kwargs is not None:
            self.module_kwargs.update(module_kwargs)
        charges1, charges2, positions, energy1, energy2 = parents
        positions.requires_grad = True
        self._index_state = IdxType.Molecules
        # self._index_state = positions._index_state
        parents = (
            charges1.main_output,
            charges2.main_output,
            positions,
            energy1.main_output,
            energy2.main_output,
        )
        super().__init__(name, parents, module=module, **kwargs)


class NACRMultiStateNode(AutoKw, SingleNode):
    """
    Compute the non-adiabatic coupling vector multiplied by the energy difference
    between all pairs of states.
    """

    _input_names = "charges", "coordinates", "energies"
    _auto_module_class = excited_layers.NACRMultiState

    def __init__(self, name, parents, module="auto", module_kwargs=None, **kwargs):
        """Automatically build the node for calculating NACR * ΔE between all pairs of
        states.

        :param name: name of the node
        :type name: str
        :param parents: parents of the NACR node in the sequence of (charges, \
                positions, energies)
        :type parents: Tuple
        :param module: _description_, defaults to "auto"
        :type module: str, optional
        :param module_kwargs: keyword arguments passed to the corresponding layer,
            defaults to None
        :type module_kwargs: dict, optional
        """

        self.module_kwargs = {}
        if module_kwargs is not None:
            self.module_kwargs.update(module_kwargs)
        charges, positions, energies = parents
        positions.requires_grad = True
        self._index_state = IdxType.Molecules
        # self._index_state = positions._index_state
        parents = (
            charges.main_output,
            positions,
            energies.main_output,
        )
        super().__init__(name, parents, module=module, **kwargs)


class LocalEnergyNode(Energies, ExpandParents, HAtomRegressor, MultiNode):
    """
    Predict a localized energy, with contributions from implicitly computed atoms.
    """

    _input_names = "hier_features", "mol_index", "atom index", "n_molecules", "n_atoms_max"
    _output_names = "mol_energy", "atom_energy", "atom_preenergy", "atom_probabilities", "atom_propensities"
    _main_output = "mol_energy"
    _output_index_states = IdxType.Molecules, IdxType.Atoms, IdxType.Atoms, IdxType.Atoms, IdxType.Atoms
    _auto_module_class = excited_layers.LocalEnergy

    @_parent_expander.match(Network)
    def expansion0(self, net, *, purpose, **kwargs):
        pdindexer = find_unique_relative(net, AtomIndexer, why_desc=purpose)
        return net, pdindexer

    @_parent_expander.match(Network, AtomIndexer)
    def expansion1(self, net, pdindexer, **kwargs):
        return net, pdindexer.mol_index, pdindexer.atom_index, pdindexer.n_molecules, pdindexer.n_atoms_max

    _parent_expander.assertlen(5)

    def __init__(self, name, parents, first_is_interacting=False, module="auto", **kwargs):
        parents = self.expand_parents(parents)
        self.module_kwargs = {"first_is_interacting": first_is_interacting}
        super().__init__(name, parents, module=module, **kwargs)

    def auto_module(self):
        network = find_unique_relative(self, Network).torch_module
        return self._auto_module_class(network.feature_sizes, **self.module_kwargs)


def _mae_with_phases(predict: torch.Tensor, true: torch.Tensor):
    """MAE with phases

    :param predict: predicted values
    :type predict: torch.Tensor
    :param true: true values
    :type true: torch.Tensor
    :return: MAE with phases
    :rtype: torch.Tensor
    """

    errors = torch.minimum(
        torch.linalg.norm(true - predict, ord=1, dim=-1),
        torch.linalg.norm(true + predict, ord=1, dim=-1),
    )
    # errors = absolute_errors(predict, true)
    return torch.sum(errors) / predict.numel()


def _mse_with_phases(predict: torch.Tensor, true: torch.Tensor):
    """MSE with phases

    :param predict: predicted values
    :type predict: torch.Tensor
    :param true: true values
    :type true: torch.Tensor
    :return: MSE with phases
    :rtype: torch.Tensor
    """

    errors = torch.minimum(
        torch.linalg.norm(true - predict, dim=-1),
        torch.linalg.norm(true + predict, dim=-1),
    )
    # errors = absolute_errors(predict, true) ** 2
    return torch.sum(errors**2) / predict.numel()


class MAEPhaseLoss(_BaseCompareLoss, op=_mae_with_phases):
    pass


class MSEPhaseLoss(_BaseCompareLoss, op=_mse_with_phases):
    pass



"""
checkpoint and state generation
"""

from typing import Tuple, Union

import torch

from ..databases import Database
from ..databases.restarter import Restartable
from ..graphs import GraphModule
from ..tools import device_fallback
from .assembly import TrainingModules
from .controllers import PatienceController
from .device import set_devices
from .metric_tracker import MetricTracker

DEFAULT_STRUCTURE_FNAME = "experiment_structure.pt"


def create_state(
    model: GraphModule,
    controller: PatienceController,
    metric_tracker: MetricTracker,
) -> dict:
    """Create an experiment state dictionary.

    :param model: current model
    :param controller: patience controller
    :param metric_tracker: current metrics
    :return: dictionary containing experiment state.
    :rtype: dict
    """
    return {
        "model": model.state_dict(),
        "controller": controller.state_dict(),
        "metric_tracker": metric_tracker,
        "torch_rng_state": torch.random.get_rng_state(),
    }


def create_structure_file(
    training_modules: TrainingModules,
    database: Database,
    controller: PatienceController,
    fname=DEFAULT_STRUCTURE_FNAME,
) -> None:
    """
    Save an experiment structure. (i.e. full model, not just state_dict).

    :param training_modules: contains model, controller, and loss
    :param database: database for training
    :param controller: patience controller
    :param fname: filename to save the checkpoint

    :return: None
    """
    structure = {
        "training_modules": training_modules,
        "controller": controller,
    }
    if isinstance(database, Restartable):
        structure["database"] = database.restarter

    with open(fname, "wb") as pfile:
        torch.save(structure, pfile)


def restore_checkpoint(structure: dict, state: dict, restore_db=True) -> dict:
    """

    :param structure: experiment structure object
    :param state: experiment state object
    :param restore_db: Attempt to restore database (true/false)

    :return: experiment structure
    """

    structure["training_modules"][0].load_state_dict(state["model"])
    structure["controller"].load_state_dict(state["controller"])

    if "database" in structure and restore_db:
        structure["database"] = structure["database"].attempt_reload()

    structure["metric_tracker"] = state["metric_tracker"]
    torch.random.set_rng_state(state["torch_rng_state"])

    return structure


def check_mapping_devices(map_location, model_device):
    """
    Check options for restarting across devices.

    :param map_location: device mapping argument for torch.load.
    :param model_device: automatically handle device mapping.
    :raises TypeError: if both map_location and model_device are specified
    :return: processed map_location and model_device
    """
    # if both are none, no transfer across device happens, directly pass map_location (which is None) to torch.load
    if model_device is not None:
        # if both map_location and model_device are given
        if map_location is not None:
            raise TypeError("Passing map_location explicitly and the model device are incompatible")
        if model_device == "auto":
            model_device = device_fallback()
        map_location = "cpu"
    return map_location, model_device


def load_saved_tensors(structure_fname: str, state_fname: str, **kwargs) -> Tuple[dict, dict]:
    """Load torch tensors from file.

    :param structure_fname: name of the structure file
    :param state_fname: name of the state file
    :return: loaded dictionaries of checkpoint and model parameters
    """

    with open(structure_fname, "rb") as pfile:
        structure = torch.load(pfile, **kwargs)

    with open(state_fname, "rb") as pfile:
        state = torch.load(pfile, **kwargs)
    return structure, state


def load_checkpoint(
    structure_fname: str, state_fname: str, restore_db=True, map_location=None, model_device=None, **kwargs
) -> dict:
    """
    Load checkpoint file from given filename.

    For details more information on to use this function, see :doc:`/examples/restarting`.

    :param structure_fname: name of the structure file
    :param state_fname: name of the state file
    :param restore_db: restore database or not, defaults to True
    :param map_location: device mapping argument for ``torch.load``, defaults to None
    :param model_device: automatically handle device mapping. Defaults to None, defaults to None
    :return: experiment structure
    """

    # we need keep the original map_location value for the if
    mapped, model_device = check_mapping_devices(map_location, model_device)
    kwargs["map_location"] = mapped
    structure, state = load_saved_tensors(structure_fname, state_fname, **kwargs)

    # transfer stuff back to model_device
    structure = restore_checkpoint(structure, state, restore_db=restore_db)
    # no transfer happens in either case, as the tensors are on the target devices already
    if model_device == "cpu" or map_location != None:
        evaluator = structure["training_modules"].evaluator
        # model_device can be None, so we have to determine the device for a tensor
        evaluator.model_device = next(evaluator.model.parameters()).device
    # if map_location is not set and model_device is set
    elif model_device != None:
        training_modules = structure["training_modules"]
        optimizer = structure["controller"].optimizer
        model, loss, evaluator = training_modules
        model, evaluator, optimizer = set_devices(model, loss, evaluator, optimizer, model_device)
    # if neither map_location nor model_device is set, directly return
    return structure


def load_checkpoint_from_cwd(map_location=None, model_device=None, **kwargs) -> dict:
    """
    Same as ``load_checkpoint``, but using default filenames.

    :param map_location: device mapping argument for ``torch.load``, defaults to None
    :type map_location: Union[str, dict, torch.device, Callable], optional
    :param model_device: automatically handle device mapping. Defaults to None, defaults to None
    :type model_device: Union[int, str, torch.device], optional
    :return: experiment structure
    :rtype: dict
    """
    return load_checkpoint(
        DEFAULT_STRUCTURE_FNAME, "best_checkpoint.pt", map_location=map_location, model_device=model_device, **kwargs
    )


def load_model_from_cwd(map_location=None, model_device=None, **kwargs) -> GraphModule:
    """
    Only load model from current working directory.

    :param map_location: device mapping argument for ``torch.load``, defaults to None
    :type map_location: Union[str, dict, torch.device, Callable], optional
    :param model_device: automatically handle device mapping. Defaults to None, defaults to None
    :type model_device: Union[int, str, torch.device], optional
    :return: model with reloaded parameters
    :rtype: GraphModule
    """
    mapped, model_device = check_mapping_devices(map_location, model_device)
    kwargs["map_location"] = mapped
    structure, state = load_saved_tensors("experiment_structure.pt", "best_model.pt", **kwargs)

    model = structure["training_modules"].model
    model.load_state_dict(state)
    if map_location == None and model_device != None and model_device != "cpu":
        model = model.to(model_device)
    return model



"""
Tools for converting between torch and numba-compatible arrays
"""
import functools

import numba
import torch
from .fast_convert import batch_convert_torch_to_numba

def via_numpy(func):
    """Decorator for piping a function through
    numpy arrays, and then giving the result back to torch.
    A bit of non-riguous testing showed that this adds overhead
    on the order of microseconds."""

    @functools.wraps(func)
    def wrapped(*args):
        args = [a.data.numpy() for a in args]
        result = func(*args)
        if not isinstance(result, (tuple, list)):
            return torch.as_tensor(result)
        return tuple(torch.as_tensor(r) for r in result)

    return wrapped

def _numba_gpu_not_found(*args,**kwargs):
    raise RuntimeError("Error: Numba not configured to run on GPU.\n"
                       "numba.cuda.is_available() returned False; numba was not able to find a GPU.\n"
                       "Verify that your numba installation is able to find cuda toolkit, as this \n"
                       "error condition likely indicates that torch can find the GPU, but numba can't.\n"
                       "Alternatively, disable custom kernels.")

class NumbaCompatibleTensorFunction:
    def __init__(self):
        if numba.cuda.is_available():
            self.kernel64 = self.make_kernel(numba.float64)
            self.kernel32 = self.make_kernel(numba.float32)
        else:
            self.kernel64 = _numba_gpu_not_found
            self.kernel32 = _numba_gpu_not_found

    def __call__(self, *args, **kwargs):
        shapes = [x.shape for x in args]
        dev = args[0].device

        if dev.type == "cpu":
            return self.cpu_kernel(*args)
        else:  # GPU
            dtype = args[0].dtype
            launch_bounds = self.launch_bounds(*shapes)
            out = torch.zeros(self.out_shape(*shapes), device=dev, dtype=dtype)
            args = *args, out
            args = [a.detach() for a in args]
            with numba.cuda.gpus[dev.index]:
                numba_args = batch_convert_torch_to_numba(*args)
                if dtype == torch.float64:
                    self.kernel64[launch_bounds](*numba_args)
                elif dtype == torch.float32:
                    self.kernel32[launch_bounds](*numba_args)
                else:
                    raise ValueError("Bad dtype: {}".format(dtype))
        return args[-1]

    def make_kernel(self):
        return NotImplemented

    def out_shape(self, *shapes):
        return NotImplemented

    def launch_bounds(self, *shapes):
        return NotImplemented

    def cpu_kernel(self):
        return NotImplemented



"""

``hippynn`` currently has interfaces to the following other codes:

1. ``ase`` for atomic simulation
2. ``pyseqm`` for combined ML-seqm models
3. ``schnetpack`` for usage of schnet models in ``hippynn``.

This subpackage is not available by default; you must import it explicitly.

"""



import torch
import time
import copy
from hippynn.experiment import serialization

time_stamps = [time.time()]

# callback function for update scf eps along epochs
def update_scf_eps(seqm, decay_factor=0.98, minimal=27.2114e-6):
    def func(epoch, better_model):
        if better_model:
            print("from callback update_scf_eps")
            p0 = seqm.energy.hamiltonian.eps.detach().data.item()
            p1 = p0 * decay_factor
            if p1 < minimal:
                p1 = minimal
            if p1 < p0:
                dtype = seqm.energy.hamiltonian.eps.dtype
                device = seqm.energy.hamiltonian.eps.device
                seqm.energy.hamiltonian.eps.data = torch.tensor(p1, dtype=dtype, device=device)
                print("SCF eps is updated: ", p0, "==>", p1)

    return func


def update_scf_backward_eps(seqm, decay_factor=0.96, minimal=1e-3):
    def func(epoch, better_model):
        if better_model:
            print("from callback update_scf_backward_eps")
            p0 = seqm.energy.hamiltonian.scf_backward_eps.detach().data.item()
            p1 = p0 * decay_factor
            if p1 < minimal:
                p1 = minimal
            if p1 < p0:
                dtype = seqm.energy.hamiltonian.scf_backward_eps.dtype
                device = seqm.energy.hamiltonian.scf_backward_eps.device
                seqm.energy.hamiltonian.scf_backward_eps.data = torch.tensor(p1, dtype=dtype, device=device)
                print("SCF scf_backward eps is updated: ", p0, "==>", p1)

    return func


def save_and_stop_after(
    training_modules, controller, metric_tracker, store_all_better=False, store_best=True, queue_time=[0, 0, 1, 0]
):
    def func(epoch, better_model):
        time_stamps.append(time.time())
        n_epochs = len(time_stamps) - 1
        time_epoch = [time_stamps[i + 1] - time_stamps[i] for i in range(n_epochs)]
        max_time_per_epoch = max(time_epoch)
        total_time = queue_time[0] * 24.0 * 3600.0 + queue_time[1] * 3600.0 + queue_time[2] * 60.0 + queue_time[3] * 1.0
        available_time = total_time - (time_stamps[-1] - time_stamps[0])
        if available_time < max_time_per_epoch * 2.0:
            # save before stop
            model = training_modules[0]
            with open("last_model.pt", "wb") as pfile:
                torch.save(copy.deepcopy(model.state_dict()), pfile)

            state = serialization.create_state(model, controller, metric_tracker)

            # Write the checkpoint
            with open("last_checkpoint.pt", "wb") as pfile:
                torch.save(state, pfile)

            if better_model:
                print("**** NEW BEST MODEL - Saving! ****")
                best_model = copy.deepcopy(model.state_dict())
                metric_tracker.best_model = best_model

                if store_all_better:
                    # Save a copy of every network doing better
                    # Note: epoch has already been incremented, so decrement in saving file.
                    with open(f"better_model_epoch_{epoch}.pt", "wb") as pfile:
                        torch.save(best_model, pfile)

                if store_best:
                    # Overwrite the "best model so far"
                    with open("best_model.pt", "wb") as pfile:
                        torch.save(best_model, pfile)

                    state = serialization.create_state(model, controller, metric_tracker)

                    # Write the checkpoint
                    with open("best_checkpoint.pt", "wb") as pfile:
                        torch.save(state, pfile)

            print("time for each epoch: ", time_epoch)
            print("Time remaining in sec: ", available_time)
            raise KeyboardInterrupt("time to stop")

    return func




import torch
import numba
from . import env_cupy
from .test_env_numba import Envops_tester, main
from .test_env_numba import TEST_MEGA_PARAMS, TEST_LARGE_PARAMS, TEST_MEDIUM_PARAMS, TEST_SMALL_PARAMS

if __name__ == "__main__":
    main(env_cupy.cupy_envsum,
         env_cupy.cupy_sensesum,
         env_cupy.cupy_featsum,
    )

